/*
** Module   : Joinable.java
** Abstract : Interface for query types which can be added as components to CompoundQuery
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20051113   @23385 Created initial version. Query types which
**                           can be added as components to CompoundQuery
**                           must implement this interface.
** 002 ECF 20051214   @23749 Added support for scrolling queries. Added
**                           several methods and extend the new Scrollable
**                           base interface.
** 003 ECF 20060106   @23797 Removed record retrieval methods. Methods
**                           first, last, next, previous, unique, and
**                           current are now defined higher up in this
**                           interface hierarchy.
** 004 ECF 20060216   @24719 Added client-side where clause support. Added
**                           setResults method to allow CompoundQuery to
**                           set a final results list into a component.
**                           Added resolveArgs parameter to reset method
**                           to allow conditional substitution parameter
**                           resolution.
** 005 ECF 20070504   @33408 Changed load() signature to return boolean.
** 006 ECF 20070518   @33681 Added getRecordBuffers(). Must return an
**                           array of RecordBuffer instances used by this
**                           query.
** 007 ECF 20080112   @37000 Removed restriction of dealing only with IDs.
**                           This interface now handles both DMO instances
**                           and primary key IDs.
** 008 ECF 20081105   @40310 Added registerRecordChangeListeners(). Used
**                           by CompoundQuery to ensure change listeners
**                           are registered with the change broker at the
**                           appropriate scope.
** 009 ECF 20090312   @42983 Removed getTableCount(). This method has moved to
**                           the P2JQuery interface.
** 010 ECF 20090703   @43052 API change. Modified load() signature.
** 011 ECF 20131028          Import change.
*/

package com.goldencode.p2j.persist;

import com.goldencode.p2j.persist.lock.*;

/**
 * Query types which can be added as components to <code>CompoundQuery</code>
 * must implement this interface.  It defines methods which will be called by
 * the enclosing, compound query object, which retrieve records and update
 * the state of the implementing query and its buffer(s).
 */
interface Joinable
extends Scrollable
{
   /**
    * Set a mode on the underlying buffer(s) such that setting a null record
    * into a buffer may or may not result in an error condition.  This applies
    * only to <code>FIRST</code>, <code>LAST</code>, and <code>UNIQUE</code>
    * iteration types.
    *
    * @param   errorIfNull
    *          If <code>true</code>, setting a <code>null</code> value as the
    *          current record in an underlying buffer will raise an error
    *          condition (if not in silent error mode); if <code>false</code>,
    *          this action will raise an end condition instead.
    */
   public void setErrorIfNull(boolean errorIfNull);
   
   /**
    * Force the query to retrieve full records, rather than primary keys only.
    */
   public void setFullRecords();
   
   /**
    * Set each buffer backing the query to unknown mode.
    *
    * @see  RecordBuffer#setUnknownMode()
    */
   public void setUnknownRecord();
   
   /**
    * Reset the query by clearing its state to a known default.  This is
    * used to ensure that subsequent requests for FIRST and LAST record
    * retrievals operate on a known query/buffer state.
    *
    * @param   resolveArgs
    *          <code>true</code> if query substitution arguments should be
    *          resolved as part of the reset;  <code>false</code> if existing
    *          argument values should be used for the next query execution.
    */
   public void reset(boolean resolveArgs);
   
   /**
    * Get an array of all record buffers managed by this query.
    *
    * @return  All record buffers managed by this query.
    */
   public RecordBuffer[] getRecordBuffers();
   
   /**
    * Load one or more records into their associated buffers, given an array
    * of primary key ID values, or the actual DMOs themselves.  The IDs/DMOs
    * are arranged to coincide with the tables underlying the query, from
    * left-most to right-most (in the sense of how they are joined by the
    * query).
    * <p>
    * The contract of this method is that if an implementor cannot load the
    * expected value (e.g., the record has been deleted or is otherwise no
    * longer available), it must throw {@link MissingRecordException}, after
    * setting the associated buffer into {@link RecordBuffer#setUnknownMode()
    * unknown mode}. Note that a record being locked by another session does
    * not constitute a "missing" record. Note also that this exception must
    * not be thrown until all the elements of <code>data</code> have been
    * processed to the extent possible.
    *
    * @param   data
    *          Array of primary key IDs or DMOs, one per table involved in the
    *          query.
    * @param   lockType
    *          Lock type which should override that of the query.  Set to
    *          <code>null</code> to allow query to use its current lock type.
    *
    * @throws  MissingRecordException
    *          if any record cannot be loaded, because it is no longer
    *          available (e.g., deleted, etc.).
    * @throws  PersistenceException
    *          if there is an error loading data.
    * 
    * @see     #getRow()
    */
   public void load(Object[] data, LockType lockType)
   throws MissingRecordException,
          PersistenceException;
   
   /**
    * Set the given results object into this query.
    *
    * @param   results
    *          New result set for this query.
    */
   public void setResults(Results results);
   
   /**
    * Assemble an array of primary key IDs or DMOs for the current record(s)
    * in the buffers underlying this query, from left-most to right-most (in
    * the sense of how the query joins the associated tables).  This method
    * essentially takes a lightweight snapshot of the currently loaded
    * record(s), so that this may be restored later.
    *
    * @return  Array of primary key IDs, one per table involved in the query.
    *
    * @see     #load(Object[], LockType)
    */
   public Object[] getRow();
   
   /**
    * Fetch the array of primary key IDs associated with the first result in
    * this query's result list.
    *
    * @return  Array of record IDs reflecting the first query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekFirst();
   
   /**
    * Fetch the array of primary key IDs associated with the last result in
    * this query's result list.
    *
    * @return  Array of record IDs reflecting the last query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekLast();
   
   /**
    * Register all <code>RecordChangeListener</code>s associated with this
    * query (if any), with the context-local <code>ChangeBroker</code> at the
    * indicated scope.
    * 
    * @param   scope
    *          Scope at which listeners should be registered with the change
    *          broker.
    */
   public void registerRecordChangeListeners(int scope);
}
