/*
** Module   : PresortQuery.java
** Abstract : A query whose results are preselected and sorted by a client
**            expression or field reference
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description---------------------------
** 001 ECF 20060103   @23795 Created initial version. A specialized
**                           preselect query which uses client-side record
**                           sorting to presort query results before loop
**                           iteration occurs.
** 002 ECF 20060208   @24722 Changed to support client-side where clause
**                           processing. Also restructured inner classes
**                           to take better advantage of shared Results
**                           implementations.
** 003 ECF 20060425   @25720 Added convenience constructors for single
**                           table instances. These mirror the
**                           constructors of the superclass, only without
**                           the order by (sort) string parameter.
**                           Tentatively removed existing, single table
**                           constructors, which are most likely obsolete.
** 004 ECF 20060503   @25959 Partially reworked break group processing.
**                           Added isLastOfBreakGroup implementation.
** 005 ECF 20060602   @26953 Use DMO instances instead of primary key IDs.
**                           Corresponds with changes in parent class.
** 006 ECF 20060608   @27003 Removed assembleSelectClause() method. It is
**                           no longer different than the superclass'
**                           implementation.
** 007 ECF 20060720   @28117 Backed out #005-006 (@26953, @27003). The
**                           parent has been refitted to collect primary
**                           keys instead of full records in the first
**                           pass query.
** 008 ECF 20060725   @28189 Modified createResults() signature to match
**                           changes in the superclass.
** 009 ECF 20060730   @28271 Code cleanup. Removed unused constructors.
** 010 ECF 20060803   @28400 Simplified createResults() method. Use
**                           getRecordBuffers() method instead of building
**                           up list manually.
** 011 ECF 20061027   @30753 Replaced Object with DataModelObject for DMO
**                           arguments to methods/constructors. Required
**                           for compile-time type safety.
** 012 ECF 20070225   @32228 Plug results resource leak. Clean up results
**                           resources once results have been sorted. This
**                           was causing both prepared statements and
**                           result sets to remain unclosed in the backing
**                           JDBC connection.
** 013 ECF 20070709   @34420 Fixed assembleSelectClause(). Changed
**                           parameter type to StringBuilder to properly
**                           override the same method of the parent class.
** 014 ECF 20070815   @34855 Support converted nested CAN-FIND statements.
**                           Use RecordBuffer's new API to push/pop temp
**                           record contexts.
** 015 ECF 20080125   @36996 Refactored sorting logic to Presorter class.
**                           PresortQuery now delegates all sorting work
**                           to Presorter.
** 016 SVL 20080523   @38373 Generated HQL query is ordered by <dmo>.id.
** 017 ECF 20090313   @41601 Forced full-records mode. Added overridden
**                           addComponent() method which calls
**                           setFullRecords(). Changed createResults().
** 018 SVL 20090316   @41612 Convenience constructors that take the sort
**                           string parameter were added.
** 019 SVL 20090505   @42093 More convenience constructors added.
** 020 ECF 20090618   @42769 Added executeScroll(). Overrides parent's
**                           implementation to use FORWARD_ONLY scroll mode,
**                           which may permit certain JDBC drivers to use
**                           server-side cursors.
** 021 SVL 20090723   @43345 Make this type of queries not natively
**                           scrollable, but enable scrolling by default.
**                           Use IncrementalResults instead of SortedResults
**                           for some cases.
** 022 SVL 20090730   @43460 Re-activated #021.
** 023 SVL 20111111          Make the class implement PresortDelegate.
** 024 ECF 20131028          Import change.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import org.hibernate.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;

/**
 * A specialized preselect query which uses one or more instances of
 * <code>Resolvable</code> provided by business application code to sort the
 * result set returned by the database.  This class supports the Progress
 * semantics of "client-side" results sorting and break groups.  Multiple
 * levels of nested break groups are supported.
 * <p>
 * Sorting criteria is specified using the <code>addSortCriterion</code>
 * method.  The order in which criteria are added is significant;  the first
 * criterion added dictates the coarsest level of sorting.  Each criterion
 * added subsequently only sorts rows within that criterion added immediately
 * previous to it.
 */
public final class PresortQuery
extends PreselectQuery
implements Presortable,
           PresortDelegate
{
   /** Helper which sorts preselected, unsorted results */
   private final Presorter sorter;
   
   /**
    * Constructor which is typically used for a multi-table query. Details
    * specific to each table component of the query are added via one of the
    * <code>addComponent</code> method variants.
    */
   public PresortQuery()
   {
      super(null);
      
      sorter = new Presorter(this);
      setScrolling();
   }
   
   /**
    * Constructor which is typically used for a multi-table query. Details
    * specific to each table component of the query are added via one of the
    * <code>addComponent</code> method variants.
    * <p>
    * This variant is used by code which is converted from an OPEN QUERY
    * statement, which permits advancing off the end of a result set without
    * raising a <code>QueryOffEndException</code>.
    *
    * @param   lenientOffEnd
    *          <code>true</code> if advancing past either end of the result
    *          set should NOT raise a <code>QueryOffEndException</code>;
    *          <code>false</code> if doing so should raise the exception.
    */
   public PresortQuery(boolean lenientOffEnd)
   {
      super(lenientOffEnd, null);
      
      sorter = new Presorter(this);
      setScrolling();
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr)
   {
      this(dmo, where, whereExpr, (DataModelObject) null, null, LockType.SHARE);
   }

   /**
    * Convenience constructor designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          Sort clause that corresponds the index which defines the sort
    *          order for this query. May be <code>null</code>.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       String sort)
   {
      this(dmo, where, whereExpr, sort, null, null, LockType.SHARE);
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       DataModelObject inverse)
   {
      this(dmo, where, whereExpr, inverse, null, LockType.SHARE);
   }

   /**
    * Convenience constructor designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          Sort clause that corresponds the index which defines the sort
    *          order for this query. May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       String sort,
                       DataModelObject inverse)
   {
      this(dmo, where, whereExpr, sort, inverse, null, LockType.SHARE);
   }

   /**
    * Convenience constructor designed for a single-table preselect query.
    * A share lock will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       Object[] args)
   {
      this(dmo, where, whereExpr, (DataModelObject) null, args, LockType.SHARE);
   }

   /**
    * Convenience constructor designed for a single-table preselect query.
    * A share lock will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          Sort clause that corresponds the index which defines the sort
    *          order for this query. May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       String sort,
                       Object[] args)
   {
      this(dmo, where, whereExpr, sort, null, args, LockType.SHARE);
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    * A share lock will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       DataModelObject inverse,
                       Object[] args)
   {
      this(dmo, where, whereExpr, inverse, args, LockType.SHARE);
   }

   /**
    * Convenience constructor designed for a single-table preselect query.
    * A share lock will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          Sort clause that corresponds the index which defines the sort
    *          order for this query. May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       String sort,
                       DataModelObject inverse,
                       Object[] args)
   {
      this(dmo, where, whereExpr, sort, inverse, args, LockType.SHARE);
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       LockType lockType)
   {
      this(dmo, where, whereExpr, (DataModelObject) null, null, lockType);
   }

   /**
    * Convenience constructor designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          Sort clause that corresponds the index which defines the sort
    *          order for this query. May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       String sort,
                       LockType lockType)
   {
      this(dmo, where, whereExpr, sort, null, null, lockType);
   }
   
   /**
    * Convenience constructor designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       DataModelObject inverse,
                       LockType lockType)
   {
      this(dmo, where, whereExpr, inverse, null, lockType);
   }

   /**
    * Convenience constructor designed for a single-table preselect query.
    * The where clause must expect no substitution parameters.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          Sort clause that corresponds the index which defines the sort
    *          order for this query. May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       String sort,
                       DataModelObject inverse,
                       LockType lockType)
   {
      this(dmo, where, whereExpr, sort, inverse, null, lockType);
   }

   /**
    * Convenience constructor designed for a single-table preselect query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       Object[] args,
                       LockType lockType)
   {
      this(dmo, where, whereExpr,(DataModelObject) null, args, lockType);
   }

   /**
    * Convenience constructor designed for a single-table preselect query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          Sort clause that corresponds the index which defines the sort
    *          order for this query. May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       String sort,
                       Object[] args,
                       LockType lockType)
   {
      this(dmo, where, whereExpr, sort, null, args, lockType);
   }

   /**
    * Convenience constructor designed for a single-table preselect query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       DataModelObject inverse,
                       Object[] args,
                       LockType lockType)
   {
      this(dmo, where, whereExpr, null, inverse, args, lockType);
   }

   /**
    * Convenience constructor designed for a single-table preselect query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          Sort clause that corresponds the index which defines the sort
    *          order for this query. May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public PresortQuery(DataModelObject dmo,
                       String where,
                       WhereExpression whereExpr,
                       String sort,
                       DataModelObject inverse,
                       Object[] args,
                       LockType lockType)
   {
      super(dmo, where, whereExpr, sort, inverse, args, lockType);

      sorter = new Presorter(this);
      setScrolling();
   }

   /**
    * Add a single sort criterion, in the form of a <code>Resolvable</code>
    * object which is resolved at sort time.  It is the value of the resolved
    * result which is used for the sort.
    * <p>
    * The sort is ascending, from lowest resolved value to highest.
    *
    * @param   sort
    *          A resolvable object whose resolved result is used for sorting.
    */
   public void addSortCriterion(Resolvable sort)
   {
      sorter.addSortCriterion(sort);
   }
   
   /**
    * Add a single sort criterion, in the form of a <code>Resolvable</code>
    * object which is resolved at runtime.  It is the value of the resolved
    * result which is used for the sort.
    *
    * @param   descending
    *          if <code>true</code>, the sort is descending, from highest
    *          resolved value to lowest;  otherwise, the sort is ascending.
    *
    * @param   sort
    *          A resolvable object whose resolved result is used for sorting.
    */
   public void addSortCriterion(Resolvable sort, boolean descending)
   {
      sorter.addSortCriterion(sort, descending);
   }
   
   /**
    * Enable all sort criteria set for this query to act as break group
    * categories.  When the query's results are presorted, the contents of
    * each break group are determined.
    */
   public void enableBreakGroups()
   {
      sorter.enableBreakGroups();
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * first row in the presorted results list.
    *
    * @return  <code>true</code> if the current result row is the first row,
    *          else <code>false</code>.
    */
   public logical isFirst()
   {
      return sorter.isFirst();
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * first row within a break group whose category is identified by the
    * specified resolvable object.
    *
    * @param   key
    *          Break group category key.  Must be the same object reference
    *          as was specified previously when invoking one of the
    *          <code>addSortCriterion</code> method variants.
    *
    * @return  <code>true</code> if break groups are enabled and current row
    *          is first in the specified break group;
    *          <code>false</code> if break groups have not been enabled or
    *          current row is not first in the specified break group.
    */
   public logical isFirstOfGroup(Resolvable key)
   {
      return sorter.isFirstOfGroup(key);
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * last row in the presorted results list.
    *
    * @return  <code>true</code> if the current result row is the last row,
    *          else <code>false</code>.
    */
   public logical isLast()
   {
      return sorter.isLast();
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * last row within a break group whose category is identified by the
    * specified resolvable object.
    *
    * @param   key
    *          Break group category key.  Must be the same object reference
    *          as was specified previously when invoking one of the
    *          <code>addSortCriterion</code> method variants.
    *
    * @return  <code>true</code> if break groups are enabled and current row
    *          is last in the specified break group;
    *          <code>false</code> if break groups have not been enabled or
    *          current row is not last in the specified break group.
    */
   public logical isLastOfGroup(Resolvable key)
   {
      return sorter.isLastOfGroup(key);
   }
   
   /**
    * Access the current, sorted results for this query.
    *
    * @return  <code>Results</code> object used to retrieve sorted results.
    */
   public Results getSortedResults()
   {
      return getResults();
   }

   /**
    * Add a component to the query.
    *
    * @param   comp
    *          Query component to be added.
    *
    * @throws  UnsupportedOperationException
    *          if attempting an outer join (temporary);
    *          if iteration type is not <code>NEXT</code> (temporary);
    *          if the component's DMO's associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    */
   protected void addComponent(Component comp)
   {
      setFullRecords();
      
      super.addComponent(comp);
   }
   
   /**
    * Does this query support non-NEXT iteration components?
    * 
    * @return  <code>true</code>
    */
   protected boolean supportsNonNextIteration()
   {
      return true;
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * last row within a break group whose category is identified by the
    * specified resolvable object.
    *
    * @param   key
    *          Break group category key.  Must be the same object reference
    *          as was specified previously when invoking one of the
    *          <code>addSortCriterion</code> method variants.
    *
    * @return  <code>true</code> if break groups are enabled and current row
    *          is last in the specified break group;
    *          <code>false</code> if break groups have not been enabled or
    *          current row is not last in the specified break group.
    */
   protected boolean isLastOfBreakGroup(Resolvable key)
   {
      return sorter.isLastOfBreakGroup(key);
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * first row within a break group whose category is identified by the
    * specified resolvable object.
    * <p>
    * This is the backing worker method for {@link #isFirstOfGroup} and is
    * used internally by the query to manage accumulators.
    *
    * @param   key
    *          Break group category key.  Must be the same object reference
    *          as was specified previously when invoking one of the
    *          <code>addSortCriterion</code> method variants.
    *
    * @return  <code>true</code> if current row is first in a break group;
    *          <code>false</code> if it is not;  unknown value if break groups
    *          have not been enabled.
    */
   protected boolean isNewBreakGroup(Resolvable key)
   {
      return sorter.isNewBreakGroup(key);
   }
   
   /**
    * Execute the query to create a <code>ScrollableResults</code> result set.
    * <p>
    * Override the parent's implementation to use a <code>ScrollMode</code> of
    * <code>FORWARD_ONLY</code>, which may permit the use of a server-side
    * cursor by the JDBC driver.
    *
    * @param   persistence
    *          Object which is used to execute the query.
    * @param   hql
    *          HQL query string.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          within the HQL query string.
    * @param   types
    *          Query substitution parameter types.
    *
    * @return  The <code>ScrollableResults</code> object returned by the
    *          execution of the query.
    *
    * @throws  PersistenceException
    *          if there is an error executing the query.
    */
   protected ScrollableResults executeScroll(Persistence persistence,
                                             String hql,
                                             Object[] args,
                                             Type[] types)
   throws PersistenceException
   {
      return persistence.scroll(hql,
                                args,
                                types,
                                0,
                                0,
                                ScrollMode.FORWARD_ONLY);
   }
   
   /**
    * Given the scrollable results set for the query, create an object which
    * implements the <code>Results</code> interface.  It is this object which
    * will actually be used to retrieve the primary key values of records to
    * be loaded from the database.
    * <p>
    * This implementation returns a specialized {@link Results} object,
    * which sorts the results obtained from the caller, according to one or
    * more sort criterion provided by client code.
    *
    * @param   results
    *          An object used to access the records which meet this query's
    *          criteria.
    *
    * @return  <code>Results</code> object which presorts the results.
    *
    * @throws  PersistenceException
    *          if there is any error creating results
    */
   protected Results createResults(Results results)
   throws PersistenceException
   {
      if (isClientWhere())
      {
         // Need to override FilteredResults to store entire records, since
         // this is what SortedResults expects.
         results = new FilteredResults(results,
                                       Arrays.asList(getRecordBuffers()),
                                       getWhereExpressions(),
                                       false);
      }
      
      // Use incremental results for single-table queries (multi-table queries
      // do not use sort phrase) which back {{DO | REPEAT} PRESLECT |
      // FOR EACH} BREAK BY statements and which sort criteria added to the
      // Presorter matches the leading components of the query's sort phrase.
      if (getTableCount() == 1 && sorter.isBreakGroupsEnabled())
      {
         List<SortCriterion> criteria =
                   SortCriterion.parse(sort, components().next().getBuffer());

         if (sorter.sortCriteriaMatchSortPhrase(criteria))
         {
            return sorter.createIncrementalResults(results);
         }
      }
      

      return sorter.createSortedResults(results);
   }

   /**
    * Indicate whether this query type is scrolling by its nature.
    * <p>
    * Subclasses which are not natively scrolling should override this method.
    *
    * @return  <code>false</code>.
    */
   protected boolean isNativelyScrolling()
   {
      return false;
   }

   /**
    * Assemble the select clause for an HQL query.  Form of the clause is:
    * <pre>
    *    select {buffer0} [, {buffer1} [...]]
    * </pre>
    * <p>
    * This method overrides the parent's implementation to make it a full
    * query, rather than projection query which returns only the record's
    * primary key.  This is necessary to presort the results without issuing
    * a separate select to the database to retrieve each record during the
    * presort process.
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    */
   protected void assembleSelectClause(StringBuilder buf)
   {
      buf.append("select ");
      
      Iterator<Component> iter = components();
      for (int i = 0; iter.hasNext(); i++)
      {
         Component comp = iter.next();
         if (i > 0)
         {
            buf.append(", ");
         }
         buf.append(comp.getBuffer().getDMOAlias());
      }
      
      buf.append(" ");
   }

   /**
    * Assemble the <code>order by</code> clause for an HQL query.  Uses the
    * form (ordered only by primary keys):
    * <pre>
    *   order by {dmo.id asc} [, ...]
    * </pre>
    * <p>
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    *
    * @return  List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    *
    * @throws  PersistenceException
    *          never.
    */
   protected List<SortCriterion> assembleOrderByClause(StringBuilder buf)
   throws PersistenceException
   {
      buf.append("order by ");

      List<SortCriterion> sortCriteria =
         (sort == null) ? new ArrayList<SortCriterion>() :
                          super.buildSortCriteria(sort);
      
      Iterator<Component> iter = components();
      while (iter.hasNext())
      {
         RecordBuffer buffer = iter.next().getBuffer();
         String schema       = buffer.getSchema();
         Class<?> dmoIface   = buffer.getDMOInterface();
         Class<?> dmoClass   = buffer.getDMOImplementationClass();

         StringBuilder text = new StringBuilder(buffer.getDMOAlias());
         text.append(".id asc");

         SortCriterion crit = new SortCriterion(buffer,
                                                text.toString(),
                                                schema,
                                                dmoIface,
                                                dmoClass);
         sortCriteria.add(crit);
      }

      return assembleOrderByClauseNoPreamble(buf, sortCriteria);
   }
}
