/*
** Module   : QueryWrapper.java
** Abstract : Proxy for concrete query implementations
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060430   @25885 Created initial version. Merged contents of
**                           AbstractQueryWrapper, CompoundQueryWrapper,
**                           PreselectQueryWrapper, and RandomAccessQuery-
**                           Wrapper. These classes are now obsolete. The
**                           merge was required because the same define
**                           query statement in Progress can be followed
**                           by many open query statements for that query;
**                           however, each open query may open a query of
**                           a different type than the previous one. Thus,
**                           a single wrapper type which supports all
**                           query types as delegates is required.
** 002 ECF 20060504   @25970 Added safety code. A default delegate query
**                           is automatically assigned upon construction,
**                           which has safe implementations of P2JQuery
**                           methods.
** 003 ECF 20060618   @27423 Change to assign() methods. Ensure query is
**                           set to scrolling upon assignment if the
**                           scrolling flag was previously set.
** 004 ECF 20061023   @30608 Added errorIfNull logic. P2JQuery interface
**                           now requires an implementation of the
**                           setErrorIfNull() method. The value set by the
**                           caller of this method is set into the query
**                           delegate, if one exists. If a delegate has
**                           not yet been assigned to the wrapper, the
**                           setting of this state in the delegate is
**                           deferred until a delegate is assigned.
** 005 ECF 20061027   @30754 Replaced Object with DataModelObject for DMO
**                           arguments to methods/constructors. Required
**                           for compile-time type safety.
** 006 ECF 20061102   @30960 Added support for reposition notification and
**                           auto-fetch features.
** 007 ECF 20061204   @31577 Implemented new P2JQuery methods to support
**                           browse.
** 008 NVS 20070502   @33345 Assigning a query should notify all listeners
**                           so that the related browse widget gets the
**                           notification and refreshes the display.
** 009 SIY 20070502   @33429 Added notifyRepositionListeners() method.
**                           Changed assignImpl() to use this new method.
**                           Minor cleanups in docs.
** 010 ECF 20070716   @34508 Added cleanup infrastructure. Wrapper is
**                           registered with TM at construction. At end of
**                           scope, cleanup is delegated to wrapped query.
**                           This ensures delegate does not clean up too
**                           early, since delegate may be assigned at a
**                           more deeply nested scope than those at which
**                           the wrapper can be used.
** 011 SIY 20070726   @34706 Helper method for explicit notification 
**                           invocations.
** 012 ECF 20080202   @37005 Changed constructor signature. Added argument
**                           to indicate whether delegate query is a
**                           scrolling query. This allows us to avoid an
**                           ugly, anonymous initializer in converted code
**                           which does nothing more than invoke the
**                           setScrolling() method. Also, automatically
**                           invoked delegate's open() method when a
**                           PreselectQuery delegate is assigned.
** 013 ECF 20080509   @38244 Fixed assignImpl(). Removed notification to
**                           reposition listeners. This is handled from
**                           the delegate's open() method, which is now
**                           part of P2JQuery and is called when the
**                           delegate is assigned.
** 014 ECF 20080509   @38253 Fixed regression introduced by #013 (@38244).
**                           assignImpl() must not call open() for a
**                           multi-table query, since this will trigger
**                           execution of the query too early. Deprecated
**                           notifyListeners(), which should be replaced
**                           in business logic with open().
** 015 ECF 20080512   @38495 Changed setScrolling(). Do not process method
**                           body more than once.
** 016 SVL 20080604   @38533 Modified according to the new P2JQuery
**                           interface.
** 017 ECF 20080729   @39327 Added rowid support. Refactored reposition*()
**                           methods.
** 018 ECF 20081106   @40325 Added _isOffEnd(). Required by P2JQuery
**                           interface change.
** 019 CA  20081210   @40840 Implemented CLOSE QUERY statement. On close,
**                           it must notify all browse widgets to reset 
**                           their state.
** 020 CA  20081211   @40860 Fixed regression introduced by H019: the 
**                           delegate needs to be explicitly closed, not
**                           cleaned.
** 021 CA  20090227   @41389 Added WrappedResources.isUnknown method; this
**                           implementation returns false.
** 022 SIY 20090306   @41425 Close delegate if new one is assigned. This is
**                           necessary to properly notify underlying browse
**                           when they are associated with query.
** 023 ECF 20090312   @42981 Implemented getTableCount(). Supports NUM-BUFFERS
**                           query handle attribute. Changed some integer
**                           parameters to recids. Use parameterized form of
**                           WeakList.
** 024 SVL 20090730   @43446 Removed open() call from assign(). It should be
**                           called externally.
** 025 CA  20090731   @43469 Register the query for off-end listener 
**                           registration after block initialization (done
**                           only for block types FOR and FOR EACH). Added
**                           getOffEndListeners, which returns a list of all
**                           off-end listeners for all used buffers.
** 026 SVL 20091030   @44289 Added getOffEnd() function to QueryWrapper and
**                           DefaultDelegate in order to support API change
**                           into P2JQuery.
** 027 SVL 20111111          presortDelegate checks that delegate is an
**                           instance of PresortDelegate (rather than
**                           PresortQuery).
** 028 VMN 20130207          Added support for handle-based methods:
**                           GET-BUFFER-HANDLE (KW_GET_BUFH)
**                           GET-CURRENT (KW_GET_CUR),
**                           QUERY-CLOSE (KW_QRY_CLOS),
**                           REPOSITION-FORWARD (KW_REPOS_F),
**                           REPOSITION-TO-ROW (KW_REPOS_2R)
**                           REPOSITION-TO-ROWID (KW_REPOS_2I),
**                           and attributes
**                           NUM-RESULTS (KW_NUM_RES)
**                           and attributes
**                           NUM-RESULTS (KW_NUM_RES)
** 029 OM  20130212          Added implementation stubs for some dynamic methods.
** 030 VMN 20130213          Added support for attributes
**                           INDEX-INFORMATION (KW_IDX_INFO)
**                           FORWARD-ONLY (KW_FWD_ONLY)
**                           PREPARE_STRING (KW_PREP_STR)
** 031 VMN 20130219          Added support for attributes
**                           IS-OPEN (KW_IS_OPEN)
** 032 CA  20130221          Added ADM-DATA, UNIQUE-ID and chaining support.
** 033 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 034 VMN 20130306          Added addBuffer(handle bufHandle) and corrected support
**                           for handle-based methods:
**                           GET-NEXT (KW_GET_NEXT)
**                           GET-PREV (KW_GET_PREV)
**                           GET-FIRST (KW_GET_1ST)
**                           GET-LAST (KW_GET_LAST)
**                           GET-CURRENT (KW_GET_CUR)
**                           Implemented
**                           getFirst(),getNext(),getLast(),getPrevious(),getCurrent()
** 035 CA  20130311          Added bufferHandle(integer).
** 036 VMN 20130311          Implemented queryClose, queryForward,
**                           queryReposition, queryRepositionByID,
**                           Delegating methods
** 037 VMN 20130312          Corrected conversion and runtime for
**                           REPOSITION-TO-ROW (KW_REPOS_2R)
** 038 CA  20130315          Added missing INDEX-INFORMATION API.
** 039 OM  20130320          Added implementation for ResultListHandler interface.
** 040 VMN 20130321          Corrected conversion and runtime for
**                           NUM-BUFFERS attribute(KW_NUM_BUFF)
** 041 CA  20130322          Added missing queryForward(integer) API.
** 042 SVL 20130325          Added include and exclude functions.
** 043 CA  20130603          resourceDelete must return a boolean value.
** 044 CA  20130927          Removed unknown() API, as by default the resource is known. 
** 045 ECF 20130405          Moved createQuery static methods here from AbstractQuery.
**                           Cleaned up some documentation.
** 046 CA  20131024          Implemented runtime for the QUERY resource (static or dynamic). 
** 047 ECF 20131028          Import change.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.UniqueIdGenerator.IdKind;

/**
 * A delegating container implementation for {@link AbstractQuery}s.  An
 * instance of this class contains an instance of <code>AbstractQuery</code>,
 * to which it delegates the work of all of its methods.
 * <p>
 * The purpose of this class is to support queries defined in Progress using
 * the DEFINE QUERY statement.  These may be defined once, but opened and
 * closed multiple times.  In the P2J implementation, a new instance of
 * <code>AbstractQuery</code> is constructed each time a query is "opened"
 * (i.e., an OPEN QUERY was issued in the pre-conversion code).  Since a
 * reference to the query may be held by many routines, the query's internal
 * state must be updated when the query is opened or re-opened.  This is
 * accomplished by sharing a reference to an instance of this class, which
 * always holds the most recently constructed, underlying query.  Thus, since
 * all routines are accessing the query via this container (i.e., essentially
 * by proxy), they will always be accessing the correct, underlying query.
 * When a new query is opened, the new <code>AbstractQuery</code> instance is
 * stored in the wrapper via one of the <code>assign</code> method variants.
 * <p>
 * The public methods of this class may only be invoked if the underlying
 * delegate query is of a type which implements a method of the same
 * signature.  Many methods are supported by {@link AbstractQuery}, which is
 * the abstract base class of all concrete query implementations.  However,
 * certain methods are implemented only by particular implementations.  These
 * will fail with an <code>IllegalStateException</code> if invoked when the
 * wrong type of delegate has been set.
 * <p>
 * When an instance of this class is first created, a no-op, default
 * implementation of <code>P2JQuery</code> is assigned immediately.  This
 * ensures that at least those methods which are defined by the
 * <code>P2JQuery</code> interface have a safe methods.  However, this does
 * not protect the caller from invoking methods which are unique to specific
 * query types, if no safe delegate has first been assigned.  This safety
 * measure is intended to protect other runtime code which generically uses
 * an instance of this class via the <code>P2JQuery</code> interface (such as
 * <code>BrowseWidge</code> in the UI).
 * <p>
 * A query wrapped by this object may need to clean up resources when the
 * wrapper goes out of scope.  A cleaner object is registered with the
 * transaction manager upon construction of the wrapper, at the next external
 * scope.  This scope assumes that a <code>QueryWrapper</code> is only
 * instantiated as an instance member of a business logic class.  This is the
 * case for all converted Progress code.
 */
@LegacyResource(resource = LegacyResource.QUERY)
public final class QueryWrapper
extends HandleChain
implements P2JQuery,
           WrappedResource,
           ADMData,
           UniqueID,
           CommonHandleChain,
           DynamicResource,
           ResultListHandler
{
   /** Default delegate to use before a real one is assigned */
   private static final P2JQuery DEFAULT_DELEGATE = new DefaultDelegate();
   
   /** Delegate query */
   private P2JQuery delegate = null;
   
   /** Reposition listeners registered with the delegate query */
   private WeakList<RepositionListener> repoListeners =
      new WeakList<RepositionListener>();
   
   /** Should delegate be a scrolling query? */
   private boolean scrolling = false;
   
   /** Should delegate's failed first/last/unique retrievals raise error? */
   private boolean errorIfNull = true;
   
   /** Should the next row of the delegate be fetched after a reposition? */
   private boolean fetchOnReposition = false;
   
   /** Is query associated with a UI browse widget? */
   private boolean browsed = false;

   /** Flag indicating if the query is closing. */
   private boolean closing = false;

   /** Flag indicating if the query was explicitly closed or is not open. */
   private boolean closed = true;

   /** Flag indicating if the query has been advanced (via GET NEXT) to the next record. */
   private boolean advanced = false;
   
   /** Flag indicating if this query is dynamic. */
   private boolean dynamic = false;
   
   /** Corresponds ADM-DATA attribute. */
   private final character admData = new character();

   /** Corresponds UNIQUE-ID attribute. */
   private final integer uniqueID = new integer();

   /** Corresponds FORWARD-ONLY attribute. */
   private boolean forwardOnly = false;
   
   /** Contains the prepare string set by {@link #prepare}. */
   private String prepareString = null;
   
   /** The set of buffers used by this dynamic query. */
   private List<Buffer> buffers = new ArrayList<Buffer>();
   
   /** Check if this query is valid. */
   private boolean valid = true;
   
   /** Internal reference for the query's name. */
   private final String qname;
   
   /**
    * Constructor which creates a static query and sets query scrolling mode for all delegate 
    * queries assigned to this wrapper.
    * 
    * @param   name
    *          The legacy name of this query.
    * @param   scrolling
    *          <code>true</code> to set all delegate queries assigned to this wrapper to scrolling
    *          mode, else <code>false</code> to not use scrolling mode.
    */
   public QueryWrapper(String name, boolean scrolling)
   {
      this(name, scrolling, false);
   }
   
   /**
    * Constructor which sets query scrolling mode for all delegate queries assigned to this 
    * wrapper.
    * 
    * @param   name
    *          The legacy name of this query.
    * @param   scrolling
    *          <code>true</code> to set all delegate queries assigned to this wrapper to scrolling
    *          mode, else <code>false</code> to not use scrolling mode.
    * @param   dynamic
    *          Flag identifying a dynamic query.
    */
   private QueryWrapper(String name, boolean scrolling, boolean dynamic)
   {
      this.name.assign(name);
      this.qname = name;
      this.dynamic = dynamic;
      this.uniqueID.assign(UniqueIdGenerator.getUniqueId(IdKind.QUERY));

      // Store a safe delegate until a real one is assigned.
      this.delegate = DEFAULT_DELEGATE;
      
      // Register a cleaner.  QueryWrappers are created as instance variables,
      // so we register to the next external scope, which corresponds to a
      // business logic class' execute() method.
      Finalizable cleaner = new Finalizable()
      {
         public void finished() { cleanup(); }
         public void iterate() {}
         public void retry() {}
      };
      TransactionManager.registerNextExternal(cleaner);
      
      if (scrolling)
      {
         setScrolling();
      }
      
      TransactionManager.registerOffEndQuery(this);
   }
   
   /**
    * Dynamically creates a query in the specified widget pool and assign 
    * it to a handle. 
    * This method is the equivalent of <code>CREATE QUERY</code> from 4GL.
    * 
    * @param   hQuery
    *          The handle to assign the newly created query.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic query. <code>null</code> if
    *          the unnamed pool must be used.
    */
   public static void createQuery(handle hQuery, String widgetPool)
   {
      createQuery(hQuery, (widgetPool == null ? null : new character(widgetPool)));
   }
   
   /**
    * Dynamically creates a query in the specified widget pool and assign 
    * it to a handle. 
    * This method is the equivalent of <code>CREATE QUERY</code> from 4GL.
    * 
    * @param   hQuery
    *          The handle to assign the newly created query.
    * @param   widgetPool
    *          The name of the widget pool that contains the dynamic query. <code>null</code> if
    *          the unnamed pool must be used.
    */
   public static void createQuery(handle hQuery, character widgetPool)
   {
      // validate widget pool before anything else!!!
      if (!WidgetPool.validWidgetPool(widgetPool))
      {
         return;
      }
      
      // create the query; use an empty string as its name
      QueryWrapper res = new QueryWrapper("", false, true);
      hQuery.assign(res);

      WidgetPool.addResource(res, widgetPool);
   }
   
   /**
    * Dynamically creates a query and assign it to a handle. This form of the 
    * method will create the object in the current default widget pool.
    * This method is the equivalent of <code>CREATE QUERY</code> from 4GL. 
    * 
    * @param   hQuery
    *          The handle to assign the newly created query.
    */
   public static void createQuery(handle hQuery)
   {
      createQuery(hQuery, (character) null);
   }
   
   /**
    * Conversion of DYNAMIC attribute. Returns <code>true</code> if the buffer is dynamic.
    *
    * @return   <code>true</code> if the resource is dynamic, <code>false</code> otherwise. 
    */
   @Override
   public logical dynamic()
   {
      return new logical(dynamic);
   }

   /**
    * Conversion of DYNAMIC attribute. Returns <code>true</code> if the buffer is dynamic.
    *
    * @return   <code>true</code> if the resource is dynamic, <code>false</code> otherwise. 
    */
   @Override
   public boolean _dynamic()
   {
      return dynamic;
   }
   
   /**
    * Assign a newly opened <code>CompoundQuery</code> instance as this
    * container's active delegate query.
    *
    * @param   delegate
    *          New delegate query.
    */
   public void assign(CompoundQuery delegate)
   {
      addBuffers(delegate, true);

      assignImpl(delegate);
   }
   
   /**
    * Assign a newly opened <code>PreselectQuery</code> instance as this
    * container's active delegate query.
    *
    * @param   delegate
    *          New delegate query.
    */
   public void assign(PreselectQuery delegate)
   {
      addBuffers(delegate, true);

      assignImpl(delegate);
   }
   
   /**
    * Assign a newly opened <code>RandomAccessQuery</code> instance as this
    * container's active delegate query.
    *
    * @param   delegate
    *          New delegate query.
    */
   public void assign(RandomAccessQuery delegate)
   {
      addBuffers(delegate, true);

      assignImpl(delegate);
      delegate.setLenientOffEnd(true);
   }
   
   /**
    * Ensure that any <code>AbstractQuery</code> assigned to this wrapper is
    * a scrolling query, to enable results list scrolling/repositioning.  If
    * a query is already assigned to this wrapper, invoke its
    * <code>setScrolling</code> method immediately.
    */
   public void setScrolling()
   {
      if (!scrolling)
      {
         scrolling = true;
         if (delegate != null)
         {
            getDelegate().setScrolling();
         }
      }
   }
   
   /**
    * Open the prepared query delegate.
    */
   public void open()
   {
      this.closed = false;
      this.advanced = false;

      delegate.open();
   }
   
   /**
    * Explicitly close the prepared query delegate.
    */
   public void close()
   {
      delegate.close();
      this.fetchOnReposition = true;
      this.closing = true;
      
      notifyRepositionListeners(true);
      
      // unregister the in-mem query class if the query was previously loaded
      if (delegate != null && delegate != DEFAULT_DELEGATE)
      {
         DynamicQueryHelper.unregisterQuery(delegate);
      }
      
      this.delegate = DEFAULT_DELEGATE;
      this.closing = false;
      this.closed = true;
   }

   /**
    * Get the number of tables joined by this query.
    * 
    * @return  Number of tables joined by this query.
    */
   public int getTableCount()
   {
      return delegate.getTableCount();
   }
   
   /**
    * Get all the off-end listeners associated with this query.
    * 
    * @return  the off-end listeners associated with this query.
    */
   public List<QueryOffEndListener> getOffEndListeners()
   {
      return delegate.getOffEndListeners();
   }

   /**
    * Set a mode on the underlying buffers such that setting a null record
    * into a buffer may or may not result in an error condition.  This applies
    * only to <code>FIRST</code>, <code>LAST</code>, and <code>UNIQUE</code>
    * iteration types.
    * <p>
    * If a query delegate has already been set, it is updated with the new
    * setting immediately.  The setting is remembered, such that any delegate
    * assigned after this method is invoked will have this setting updated
    * upon assignment to the query wrapper.
    *
    * @param   errorIfNull
    *          If <code>true</code>, setting a <code>null</code> value as the
    *          current record in an underlying buffer will raise an error
    *          condition (if not in silent error mode); if <code>false</code>,
    *          this action will raise an end condition instead.
    */
   public void setErrorIfNull(boolean errorIfNull)
   {
      this.errorIfNull = errorIfNull;
      if (delegate != null)
      {
         getDelegate().setErrorIfNull(errorIfNull);
      }
   }
   
   /**
    * Add a joinable query to a compound query.  If this is not the first
    * component to be added to the query, its results will be joined to the
    * immediately previous component by an inner join.  This query will be
    * iterated using the <code>NEXT</code> iteration type.
    *
    * @param   query
    *          Query which will perform the fundamental work for this
    *          component when the compound query is iterated.
    *
    * @throws  IllegalStateException
    *          if this method is invoked during an iteration cycle or if
    *          delegate is not an instance of <code>CompoundQuery</code>.
    *
    * @see     #iterate
    */
   public void addComponent(Joinable query)
   {
      addBuffers(query, false);

      compoundDelegate().addComponent(query);
   }
   
   /**
    * Add a joinable query to a compound query, and specify the nature of
    * the iteration which will be performed (first, last, next, etc.) on that
    * component, when the compound query is iterated.  If this is not the
    * first component to be added to the query, its results will be joined to
    * the immediately previous component by an inner join.
    *
    * @param   query
    *          Query which will perform the fundamental work for this
    *          component when the compound query is iterated.
    * @param   iteration
    *          A {@link QueryConstants constant} indicating the type of work
    *          to be performed upon an iteration of the compound query.  This
    *          parameter determines whether the underlying query will be asked
    *          to perform a first, last, next, previous, or unique record
    *          retrieval.
    *
    * @throws  IllegalStateException
    *          if this method is invoked during an iteration cycle or if
    *          delegate is not an instance of <code>CompoundQuery</code>.
    *
    * @see     #iterate
    */
   public void addComponent(Joinable query, int iteration)
   {
      addBuffers(query, false);

      compoundDelegate().addComponent(query, iteration);
   }
   
   /**
    * Add a joinable query to a compound query, and specify the nature of
    * the iteration which will be performed (first, last, next, etc.) on that
    * component, when the compound query is iterated.  If this is not the
    * first component to be added to the query, the type of join performed
    * with the immediately previous component is determined by the value of
    * <code>outer</code>.
    *
    * @param   query
    *          Query which will perform the fundamental work for this
    *          component when the compound query is iterated.
    * @param   iteration
    *          A {@link QueryConstants constant} indicating the type of work
    *          to be performed upon an iteration of the compound query.  This
    *          parameter determines whether the underlying query will be asked
    *          to perform a first, last, next, previous, or unique record
    *          retrieval.
    * @param   outer
    *          <code>true</code> if this component should be joined to the
    *          previously added component via a left outer join;
    *          <code>false</code> if an inner join is required.  Note that
    *          adding an inner join component will cause any previously added
    *          components added as outer joins to be reset to inner joins.
    *
    * @throws  IllegalStateException
    *          if this method is invoked during an iteration cycle or if
    *          delegate is not an instance of <code>CompoundQuery</code>.
    *
    * @see     #iterate
    */
   public void addComponent(Joinable query, int iteration, boolean outer)
   {
      addBuffers(query, false);

      compoundDelegate().addComponent(query, iteration, outer);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo, String where)
   {
      preselectDelegate().addComponent(dmo, where);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse)
   {
      preselectDelegate().addComponent(dmo, where, inverse);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo, String where, Object[] args)
   {
      preselectDelegate().addComponent(dmo, where, args);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            Object[] args)
   {
      preselectDelegate().addComponent(dmo, where, inverse, args);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            LockType lockType)
   {
      preselectDelegate().addComponent(dmo, where, lockType);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            LockType lockType)
   {
      preselectDelegate().addComponent(dmo, where, inverse, lockType);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            Object[] args,
                            LockType lockType)
   {
      preselectDelegate().addComponent(dmo, where, args, lockType);
   }
   
   /**
    * Add a component to a preselect query.  The following defaults are used:
    * <ul>
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType)
   {
      preselectDelegate().addComponent(dmo, where, inverse, args, lockType);
   }
   
   /**
    * Add a component to a preselect query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.  At
    *          this time, only <code>NEXT</code> is supported.  Future
    *          support for <code>FIRST</code>, <code>LAST</code>, and
    *          <code>UNIQUE</code> is expected.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      preselectDelegate().addComponent(dmo,
                                       where,
                                       args,
                                       lockType,
                                       iteration,
                                       outer);
   }
   
   /**
    * Add a component to a preselect query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.  At
    *          this time, only <code>NEXT</code> is supported.  Future
    *          support for <code>FIRST</code>, <code>LAST</code>, and
    *          <code>UNIQUE</code> is expected.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      preselectDelegate().addComponent(dmo,
                                       where,
                                       inverse,
                                       args,
                                       lockType,
                                       iteration,
                                       outer);
   }
   
   /**
    * Add a client-side where expression to a preselect query.  Each initial
    * query result found by the server must be accepted by this expression
    * before becoming a part of the query's final result set.
    *
    * @param   whereExpr
    *          Client-side where clause expression.  May not be
    *          <code>null</code>.
    *
    * @throws  NullPointerException
    *          if <code>whereExpr</code> is <code>null</code>.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void addWhereExpression(WhereExpression whereExpr)
   {
      preselectDelegate().addWhereExpression(whereExpr);
   }
   
   /**
    * Set the iteration type for the most recently added query component to a
    * preselect query.
    *
    * @param   iteration
    *          Retrieval mode from {@link QueryConstants}.
    *
    * @throws  IllegalStateException
    *          If no component has yet been added, or
    *          if delegate is not an instance of <code>PreselectQuery</code>.
    */
   public void setIterationType(int iteration)
   {
      preselectDelegate().setIterationType(iteration);
   }
   
   /**
    * Add a single sort criterion, in the form of a <code>Resolvable</code>
    * object which is resolved at sort time.  It is the value of the resolved
    * result which is used for the sort.
    * <p>
    * The sort is ascending, from lowest resolved value to highest.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @param   sort
    *          A resolvable object whose resolved result is used for sorting.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public void addSortCriterion(Resolvable sort)
   {
      presortDelegate().addSortCriterion(sort);
   }
   
   /**
    * Add a single sort criterion, in the form of a <code>Resolvable</code>
    * object which is resolved at runtime.  It is the value of the resolved
    * result which is used for the sort.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @param   sort
    *          A resolvable object whose resolved result is used for sorting.
    * @param   descending
    *          if <code>true</code>, the sort is descending, from highest
    *          resolved value to lowest;  otherwise, the sort is ascending.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public void addSortCriterion(Resolvable sort, boolean descending)
   {
      presortDelegate().addSortCriterion(sort, descending);
   }
   
   /**
    * Enable all sort criteria set for this query to act as break group
    * categories.  When the query's results are presorted, the contents of
    * each break group are determined.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public void enableBreakGroups()
   {
      presortDelegate().enableBreakGroups();
   }
   
   /**
    * Add an accumulator to this query.  The accumulator expects to be
    * notified whenever a new result row is accessed (during sequential row
    * access).
    * <p>
    * This implementation assumes the accumulator is not associated with a
    * break group.
    *
    * @param   accum
    *          Accumulator to be registered with this query.
    */
   public void addAccumulator(Accumulator accum)
   {
      getDelegate().addAccumulator(accum);
   }
   
   /**
    * Add an accumulator to this query.  The accumulator expects to be
    * notified whenever a new result row is accessed (during sequential row
    * access).
    * <p>
    * This implementation assumes the accumulator is not associated with a
    * break group.
    *
    * @param   accum
    *          Accumulator to be registered with this query.
    * @param   deferred
    *          If <code>true</code>, calls to {@link #iterate()} will not
    *          cause the accumulator to recalculate immediately, but instead
    *          to update the accumulator's state such that the next invocation
    *          of {@link com.goldencode.p2j.util.Accumulator#accumulate()} will
    *          trigger the actual accumulation update.  If <code>false</code>,
    *          the accumulator recalculates its internal state immediately
    */
   public void addAccumulator(Accumulator accum, boolean deferred)
   {
      getDelegate().addAccumulator(accum, deferred);
   }
   
   /**
    * Reports if this query (that is, its delegate query) is valid for use.
    *
    * @return  <code>true</code> if this query's delegate is valid for use.
    */
   public boolean valid()
   {
      return valid;
   }

   /**
    * Advance a compound query, triggering one or more of the underlying
    * queries to retrieve a new record.  This produces a virtual, composite
    * row of results which is the combination of the records retrieved by
    * each component query.
    *
    * @throws  QueryOffEndException
    *          if no more records can be retrieved by the underlying queries.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void iterate()
   {
      compoundDelegate().iterate();
   }
   
   /**
    * Navigate to the first record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters and
    * lock type.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void first()
   {
      if (!canGetStmt() || !canGetFirst())
      {
         return;
      }

      advanced = true;

      getDelegate().first();
   }
   
   /**
    * Navigate to the first record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters, but
    * override the default lock type with that provided.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void first(LockType lockType)
   {
      if (!canGetStmt() || !canGetFirst())
      {
         return;
      }

      advanced = true;
      
      getDelegate().first(lockType);
   }
   
   /**
    * Retrieve the first virtual, composite row of results for a compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void first(boolean iterating)
   {
      compoundDelegate().first(iterating);
   }
   
   /**
    * Retrieve the first virtual, composite row of results for a compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void first(boolean iterating, LockType lockType)
   {
      compoundDelegate().first(iterating, lockType);
   }
   
   /**
    * Navigate to the first record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default lock type, but override the
    * default substitution parameters with those provided.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void first(Object[] values)
   {
      randomAccessDelegate().first(values);
   }
   
   /**
    * Navigate to the first record which meets the query criteria and retrieve
    * it into its record buffer.  Override both the default substitution
    * parameters, and the default lock type with those provided.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void first(Object[] values, LockType lockType)
   {
      randomAccessDelegate().first(values, lockType);
   }
   
   /**
    * Navigate to the last record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters and
    * lock type.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void last()
   {
      if (!canGetStmt() || !canLastOrPrev())
      {
         return;
      }

      getDelegate().last();
   }
   
   /**
    * Navigate to the last record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters, but
    * override the default lock type with that provided.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void last(LockType lockType)
   {
      if (!canGetStmt() || !canLastOrPrev())
      {
         return;
      }

      getDelegate().last(lockType);
   }
   
   /**
    * Retrieve the last virtual, composite row of results for a compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void last(boolean iterating)
   {
      compoundDelegate().last(iterating);
   }
   
   /**
    * Retrieve the last virtual, composite row of results for a compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void last(boolean iterating, LockType lockType)
   {
      compoundDelegate().last(iterating, lockType);
   }
   
   /**
    * Navigate to the last record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default lock type, but override the
    * default substitution parameters with those provided.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void last(Object[] values)
   {
      randomAccessDelegate().last(values);
   }
   
   /**
    * Navigate to the last record which meets the query criteria and retrieve
    * it into its record buffer.  Override both the default substitution
    * parameters, and the default lock type with those provided.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if no record
    *          could be found when operating in standalone mode.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void last(Object[] values, LockType lockType)
   {
      randomAccessDelegate().last(values, lockType);
   }
   
   /**
    * Navigate to the next record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters and
    * lock type.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the next record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the first record
    * which meets the query criteria.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    */
   public void next()
   {
      if (!canGetStmt())
      {
         return;
      }

      advanced = true;
      
      getDelegate().next();
   }
   
   /**
    * Navigate to the next record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default substitution parameters, but
    * override the default lock type with that provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the next record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the first record
    * which meets the query criteria.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    */
   public void next(LockType lockType)
   {
      if (!canGetStmt())
      {
         return;
      }

      advanced = true;
      
      getDelegate().next(lockType);
   }
   
   /**
    * Retrieve the next virtual, composite row of results for a compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void next(boolean iterating)
   {
      compoundDelegate().next(iterating);
   }
   
   /**
    * Retrieve the next virtual, composite row of results for a compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void next(boolean iterating, LockType lockType)
   {
      compoundDelegate().next(iterating, lockType);
   }
   
   /**
    * Navigate to the next record which meets the query criteria and retrieve
    * it into its record buffer.  Use the default lock type, but override the
    * default substitution parameters with those provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the next record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the first record
    * which meets the query criteria.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void next(Object[] values)
   {
      randomAccessDelegate().next(values);
   }
   
   /**
    * Navigate to the next record which meets the query criteria and retrieve
    * it into its record buffer.  Override both the default substitution
    * parameters, and the default lock type with those provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the next record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the first record
    * which meets the query criteria.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void next(Object[] values, LockType lockType)
   {
      randomAccessDelegate().next(values, lockType);
   }
   
   /**
    * Navigate to the previous record which meets the query criteria and
    * retrieve it into its record buffer.  Use the default substitution
    * parameters and lock type.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the previous record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the last record
    * which meets the query criteria.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    */
   public void previous()
   {
      if (!canGetStmt() || !canLastOrPrev())
      {
         return;
      }

      getDelegate().previous();
   }
   
   /**
    * Navigate to the previous record which meets the query criteria and
    * retrieve it into its record buffer.  Use the default substitution
    * parameters, but override the default lock type with that provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the previous record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the last record
    * which meets the query criteria.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    */
   public void previous(LockType lockType)
   {
      if (!canGetStmt() || !canLastOrPrev())
      {
         return;
      }

      getDelegate().previous(lockType);
   }
   
   /**
    * Retrieve the previous virtual, composite row of results for a compound
    * query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void previous(boolean iterating)
   {
      compoundDelegate().previous(iterating);
   }
   
   /**
    * Retrieve the previous virtual, composite row of results for a compound
    * query, overriding the lock type to apply to each underlying query.
    * <p>
    * This is a combination of records retrieved by each component query.  If
    * the compound query involves outer joins, some of the underlying buffers
    * may be empty when this method returns.
    *
    * @param   iterating
    *          <code>true</code> if this method is being invoked within the
    *          context of an iterating loop, which can only be exited by
    *          raising an end condition;  <code>false</code> if end condition
    *          should not be raised, even if no further advancement of the
    *          composite result is possible.
    * @param   lockType
    *          Lock type which should override that of any underlying query.
    *
    * @throws  QueryOffEndException
    *          if no more results were available in iterating mode.
    * @throws  ErrorConditionException
    *          if an underlying query triggers an error.
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>CompoundQuery</code>.
    */
   public void previous(boolean iterating, LockType lockType)
   {
      compoundDelegate().previous(iterating, lockType);
   }
   
   /**
    * Navigate to the previous record which meets the query criteria and
    * retrieve it into its record buffer.  Use the default lock type, but
    * override the default substitution parameters with those provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the previous record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the last record
    * which meets the query criteria.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void previous(Object[] values)
   {
      randomAccessDelegate().previous(values);
   }
   
   /**
    * Navigate to the previous record which meets the query criteria and
    * retrieve it into its record buffer.  Override both the default
    * substitution parameters, and the default lock type with those provided.
    * <p>
    * Use the record most recently loaded into the backing buffer as the
    * reference point when determining the previous record to visit.  If no
    * record has yet been loaded into the buffer, retrieve the last record
    * which meets the query criteria.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void previous(Object[] values, LockType lockType)
   {
      randomAccessDelegate().previous(values, lockType);
   }
   
   /**
    * Reload the record most recently loaded into the backing record buffer,
    * if any.  Use the default lock type.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record reload.
    */
   public void current()
   {
      if (!canGetStmt())
      {
         return;
      }

      getDelegate().current();
   }
   
   /**
    * Reload the record most recently loaded into the backing record buffer,
    * if any.  Override the default lock type.
    *
    * @param   lockType
    *          New lock type to apply to the record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record reload, or the lock could
    *          not be transitioned to the requested type.
    */
   public void current(LockType lockType)
   {
      if (!canGetStmt())
      {
         return;
      }

      getDelegate().current(lockType);
   }
   
   /**
    * Navigate to a particular record which meets the query criteria and
    * retrieve it into its record buffer.  No more than one record may match
    * the criteria.  Use the default substitution parameters and lock type.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if more than
    *          one record is found which match the criteria.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void unique()
   {
      getDelegate().unique();
   }
   
   /**
    * Navigate to a particular record which meets the query criteria and
    * retrieve it into its record buffer.  No more than one record may match
    * the criteria.  Use the default substitution parameters, but override
    * the default lock type with that provided.
    *
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if more than
    *          one record is found which match the criteria.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    */
   public void unique(LockType lockType)
   {
      getDelegate().unique(lockType);
   }
   
   /**
    * Navigate to a particular record which meets the query criteria and
    * retrieve it into its record buffer.  No more than one record may match
    * the criteria.  Use the default lock type, but override the default
    * substitution parameters with those provided.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if more than
    *          one record is found which match the criteria.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void unique(Object[] values)
   {
      randomAccessDelegate().unique(values);
   }
   
   /**
    * Navigate to a particular record which meets the query criteria and
    * retrieve it into its record buffer.  No more than one record may match
    * the criteria.  Override both the default substitution parameters, and
    * the default lock type with those provided.
    * <p>
    * The delegate must be an instance of <code>RandomAccessQuery</code>
    *
    * @param   values
    *          Substitution values to use when issuing HQL queries.
    * @param   lockType
    *          Lock type to acquire for the retrieved record.
    *
    * @throws  ErrorConditionException
    *          if an error occurred during record retrieval or if more than
    *          one record is found which match the criteria.
    * @throws  QueryOffEndException
    *          if no record could be found when operating in compound query
    *          mode.
    * @throws  IllegalStateException
    *          if delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   public void unique(Object[] values, LockType lockType)
   {
      randomAccessDelegate().unique(values, lockType);
   }
   
   /**
    * Is this query set to fetch the next record after a reposition?
    *
    * @return  <code>true</code> if set to fetch after reposition, else
    *          <code>false</code>.
    */
   public boolean isFetchOnReposition()
   {
      return fetchOnReposition;
   }
   
   /**
    * Set this query to automatically fetch the next record after a
    * reposition, or turn this feature off.
    *
    * @param   enable
    *          <code>true</code> to enable automatic fetching;
    *          <code>false</code> to disable.
    */
   public void setFetchOnReposition(boolean enable)
   {
      fetchOnReposition = enable;
      if (delegate != null)
      {
         getDelegate().setFetchOnReposition(enable);
      }
   }
   
   /**
    * Register a listener to receive notifications reposition events for this
    * query.
    *
    * @param   listener
    *          Reposition listener to be registered.
    */
   public void addRepositionListener(RepositionListener listener)
   {
      repoListeners.add(listener);
      if (delegate != null)
      {
         getDelegate().addRepositionListener(listener);
      }
   }
   
   /**
    * Indicate to this query that whether it is associated with a browse
    * widget or not.
    *
    * @param   browsed
    *          <code>true</code> to indicate the query has an associated
    *          browse widget, else <code>false</code>.
    */
   public void setBrowsed(boolean browsed)
   {
      this.browsed = browsed;
      if (delegate != null)
      {
         getDelegate().setBrowsed(browsed);
      }
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified primary key ID.  A
    * request to retrieve the previous result will retrieve the result
    * immediately previous to this, if any, in the results list.
    *
    * @param   id
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * 
    * @throws  ErrorConditionException
    *          if the specified ID represents the unknown value.
    */
   public void repositionByID(recid id)
   {
      getDelegate().repositionByID(id);
   }
   
   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    *
    * @param   id1
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * @param   joinIDs
    *          All remaining primary key IDs, if any, arranged from left to
    *          right to coincide with the records being joined by the
    *          underlying query.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query;
    *          if any of the specified IDs represents the unknown value.
    */
   public void repositionByID(rowid id1, rowid... joinIDs)
   {
      getDelegate().repositionByID(id1, joinIDs);
   }
   
   /**
    * Reposition the cursor to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the result at the specified row,
    * and a call to retrieve the previous result will get the result before
    * the new position.
    *
    * @param   row
    *          1-based index of the target position.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query;
    *          if <code>row</code> represents unknown value.
    */
   public void reposition(NumberType row)
   {
      getDelegate().reposition(row);
   }
   
   /**
    * Reposition the query to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the record after the new position,
    * and a call to retrieve the previous record will get the record before
    * the new position.
    *
    * @param   row
    *          1-based index of the target position.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public void reposition(int row)
   {
      getDelegate().reposition(row);
   }
   
   /**
    * Advance the current cursor position forward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it ahead <i>one and a half positions</i>, to just
    *       beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it ahead <i>one position</i>, to just beyond the
    *       following result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query;
    *          if <code>rows</code> represents the unknown value.
    */
   public void forward(NumberType rows)
   {
      getDelegate().forward(rows);
   }
   
   /**
    * Advance the current cursor position forward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it ahead <i>one and a half positions</i>, to just
    *       beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it ahead <i>one position</i>, to just beyond the
    *       following result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public void forward(int rows)
   {
      getDelegate().forward(rows);
   }
   
   /**
    * Move the current cursor position backward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it back <i>one half position</i>, to just
    *       before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it back <i>one position</i>, to just before the
    *       current result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query;
    *          if <code>rows</code> represents the unknown value.
    */
   public void backward(NumberType rows)
   {
      getDelegate().backward(rows);
   }
   
   /**
    * Move the current cursor position backward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it back <i>one half position</i>, to just
    *       before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it back <i>one position</i>, to just before the
    *       current result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public void backward(int rows)
   {
      getDelegate().backward(rows);
   }
   
   /**
    * Get the <b><i>current</i></b> size of the cached results list.  However,
    * since this cache is built progressively as the query is scrolled,
    * repositioned, and naturally navigated, this <b>does not</b> necessarily
    * represent the full count of results which may satisfy the query.
    *
    * @return  Current results list size or unknown value if the query is not
    *          open.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public integer size()
   {
      return getDelegate().size();
   }
   
   /**
    * Progress-compatible function that returns the 1-based index of the
    * current row in the result set. This is the index of the entry at or
    * before which the cursor currently is positioned.
    *
    * @return  Unknown value if the query is not open or result set is empty.
    *          1 if the query is positioned before the first record.
    *          A value 1 greater than the number of rows in the query result
    *          set if the query is positioned beyond the last record.
    *          Otherwise returns the number of the current row in the query
    *          result set.
    */
   public integer currentRow()
   {
      return getDelegate().currentRow();
   }
   
   /**
    * Get the 1-based index of the current row in the result set.  This is
    * the index of the entry at or before which the cursor currently is
    * positioned.
    *
    * @return  Current row or the unknown value if the current row cannot be
    *          determined.  The current row cannot be determined if the
    *          results list is empty, or if the cursor currently is in
    *          backward scroll mode.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public integer currentRowImpl()
   {
      return getDelegate().currentRowImpl();
   }
   
   /**
    * Indicate whether the cursor has run off one or the other end of the
    * associated query's list of available results.  If <code>true</code>,
    * this indicates that the query cannot produce any more results in the
    * current scroll direction.
    *
    * @return  <code>true</code> if the cursor is off either end of its
    *          query's results list.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public logical isOffEnd()
   {
      return (prepareString != null || dynamic) && closed ? new logical()
                                                          : getDelegate().isOffEnd();
   }
   
   /**
    * Indicate whether the cursor has run off one or the other end of the
    * associated query's list of available results.  If <code>true</code>,
    * this indicates that the query cannot produce any more results in the
    * current scroll direction.
    *
    * @return  <code>true</code> if the cursor is off either end of its
    *          query's results list.
    *
    * @throws  ErrorConditionException
    *          if this query is not a scrolling query.
    */
   public boolean _isOffEnd()
   {
      return getDelegate()._isOffEnd();
   }

   /**
    * Get the off-end status of the query, which indicates whether the query
    * has run off either end of its results.
    *
    * @return  One of the {@link OffEnd} constants <code>NONE</code>,
    *          <code>FRONT</code>, or <code>BACK</code>.
    */
   public OffEnd getOffEnd()
   {
      return getDelegate().getOffEnd();
   }

   /**
    * Determine whether the query result row currently being visited is the
    * first row in the presorted results list.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @return  <code>true</code> if the current result row is the first row,
    *          else <code>false</code>.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public logical isFirst()
   {
      return presortDelegate().isFirst();
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * first row within a break group whose category is identified by the
    * specified resolvable object.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @param   key
    *          Break group category key.  Must be the same object reference
    *          as was specified previously when invoking one of the
    *          <code>addSortCriterion</code> method variants.
    *
    * @return  <code>true</code> if current row is first in a break group;
    *          <code>false</code> if it is not;  unknown value if break groups
    *          have not been enabled.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public logical isFirstOfGroup(Resolvable key)
   {
      return presortDelegate().isFirstOfGroup(key);
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * last row in the presorted results list.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @return  <code>true</code> if the current result row is the last row,
    *          else <code>false</code>.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public logical isLast()
   {
      return presortDelegate().isLast();
   }
   
   /**
    * Determine whether the query result row currently being visited is the
    * first row within a break group whose category is identified by the
    * specified resolvable object.
    * <p>
    * The delegate must be an instance of <code>PresortDelegate</code>
    *
    * @param   key
    *          Break group category key.  Must be the same object reference
    *          as was specified previously when invoking one of the
    *          <code>addSortCriterion</code> method variants.
    *
    * @return  <code>true</code> if current row is first in a break group;
    *          <code>false</code> if it is not;  unknown value if break groups
    *          have not been enabled.
    *
    * @throws  IllegalStateException
    *          if delegate is not an instance of <code>PresortDelegate</code>.
    */
   public logical isLastOfGroup(Resolvable key)
   {
      return presortDelegate().isLastOfGroup(key);
   }
   
   /**
    * Indicate whether this query preselects all of its results.
    *
    * @return  <code>true</code> if the query preselects its results;
    *          <code>false</code> if the query's results are dynamically
    *          navigated.
    */
   public boolean isPreselect()
   {
      return getDelegate().isPreselect();
   }
   
   /**
    * Notify all registered listeners that query is repositioned or closed.
    * 
    * @param   closed
    *          A <code>true</code> value will inform all listeners that the
    *          query was explicitly closed.
    */
   public void notifyRepositionListeners(boolean closed)
   {
      delegate.notifyRepositionListeners(closed);
   }
   
   /**
    * Notify all registered listeners that query is repositioned in case if
    * query is attached to browse widget.
    * 
    * @deprecated Use #open() instead.
    */
   public void notifyListeners()
   {
      open();
   }
   
   /**
    * Indicate whether this query stands alone, or is the component of
    * another query, or is wrapped by a query wrapper.
    * 
    * @return  <code>true</code>;  query wrappers may not themselves be
    *          wrapped.
    */
   public boolean isStandalone()
   {
      return true;
   }
   
   /**
    * Change the standalone nature of this query.  A query is standalone by
    * default, but can be changed to non-standalone by the wrapper or
    * container which contains it.
    * <p>
    * Instances of <code>QueryWrapper</code> may not be wrapped in other
    * queries or query wrappers, so this implementation throws an exception
    * if this method is invoked with a parameter of <code>true</code>.
    * 
    * @param   standalone
    *          <code>true</code> to set query as standalone;
    *          <code>false</code> to set query as contained.
    *
    * @throws  UnsupportedOperationException
    *          if <code>standalone</code> is set to <code>true</code>.
    */
   public void setStandalone(boolean standalone)
   {
      if (!standalone)
      {
         throw new UnsupportedOperationException(
            "Query wrapper must be standalone");
      }
   }
   
   /**
    * Delegate cleanup processing to the query wrapped by this object.
    * 
    * @see  P2JQuery#cleanup()
    */
   public void cleanup()
   {
      if (!closed)
      {
         delegate.cleanup();
      }
   }

   /**
    * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
    *
    * Returns a handle of a particular buffer of a query object.
    *
    * @param   bufferName
    *          buffer-name. A CHARACTER expression that evaluates to the name
    *          of a buffer in the query or ProDataSet object.
    *
    * @return buffer handle
    */
   public handle bufferHandle(String bufferName)
   {
      return bufferHandle(new character(bufferName));
   }

   /**
    * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
    *
    * Returns a handle of a particular buffer of a query object.
    *
    * @param   bufferName
    *          buffer-name. A CHARACTER expression that evaluates to the name
    *          of a buffer in the query or ProDataSet object.
    *
    * @return buffer handle
    */
   public handle bufferHandle(character bufferName)
   {
      for (Buffer buf : buffers)
      {
         if (CompareOps._isEqual(bufferName, ((CommonHandleChain) buf).getName()))
         {
            return new handle(buf);
         }
      }
      
      String errMsg = "Query method argument must be the name of a query buffer or an integer " +
                      "from 1 to %d for query %s";
      errMsg = String.format(errMsg, numBuffers().intValue(), qname);
      
      ErrorManager.recordOrShowError(7361, errMsg, false, false, false);
      
      return new handle();
   }

   /**
    * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
    *
    * Returns a handle of a particular buffer of a query object.
    *
    * @param   bufferSequenceNumber
    *          buffer-sequence-number. An INTEGER that represents the sequence
    *          number of the desired buffer.
    *
    * @return buffer handle
    */
   public handle bufferHandle(int bufferSequenceNumber)
   {
      return bufferHandle(new integer(bufferSequenceNumber));
   }

   /**
    * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
    *
    * Returns a handle of a particular buffer of a query object.
    *
    * @param   bufferSequenceNumber
    *          buffer-sequence-number. An INTEGER that represents the sequence
    *          number of the desired buffer.
    *
    * @return buffer handle
    */
   public handle bufferHandle(integer bufferSequenceNumber)
   {
      int idx = bufferSequenceNumber.isUnknown() ? 0 : bufferSequenceNumber.intValue();
      
      if (idx >= 1 && idx <= buffers.size())
      {
         return new handle(buffers.get(idx - 1));
      }
      
      String errMsg = "Query method argument must be the name of a query buffer or an integer " +
                      "from 1 to %d for query %s";
      errMsg = String.format(errMsg, numBuffers().intValue(), qname);

      ErrorManager.recordOrShowError(7361, errMsg, false, false, false);

      return new handle();
   }

   /**
    * Conversion of QUERY-CLOSE() method (KW_QRY_CLOS).
    *
    * Closes a query object.
    *
    * @return  <code>true</code> if QUERY-CLOSE() succeeds,
    *          else <code>false</code>.
    */
   public logical queryClose()
   {
      boolean success = true;

      try
      {
         close();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }

      return new logical(success);
   }

   /**
    * Conversion of REPOSITION-FORWARD() method (KW_REPOS_F).
    *
    * Moves a query object's result list pointer forward a particular number
    * of rows.
    *
    * @return  <code>true</code> if REPOSITION-FORWARD() succeeds,
    *          else <code>false</code>.
    */
   public logical queryForward(int rows)
   {
      return queryForward(new integer(rows));
   }

   /**
    * Conversion of REPOSITION-FORWARD() method (KW_REPOS_F).
    *
    * Moves a query object's result list pointer forward a particular number
    * of rows.
    *
    * @return  <code>true</code> if REPOSITION-FORWARD() succeeds,
    *          else <code>false</code>.
    */
   public logical queryForward(integer rows)
   {
      if (rows.isUnknown())
      {
         // TODO: what error should be shown?
         return new logical(false);
      }

      boolean success = true;

      try
      {
         forward(rows);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }

      return new logical(success);
   }

   /**
    * Conversion of REPOSITION-TO-ROW() method (KW_REPOS_2R).
    *
    * Moves a query object's result list pointer to the row corresponding to the specified
    * sequence number.
    *
    * @param   n
    *          An integer expression representing the sequence number.
    *
    * @return  <code>true</code> if REPOSITION-TO-ROW() succeeds,
    *          else <code>false</code>.
    */
   public logical queryReposition(int n)
   {
      return queryReposition(new integer(n));
   }

   /**
    * Conversion of REPOSITION-TO-ROW() method (KW_REPOS_2R).
    *
    * Moves a query object's result list pointer to the row corresponding to the specified
    * sequence number.
    *
    * @param   n
    *          An integer expression representing the sequence number.
    *
    * @return  <code>true</code> if REPOSITION-TO-ROW() succeeds,
    *          else <code>false</code>.
    */
   public logical queryReposition(NumberType n)
   {
      boolean success = true;

      try
      {
         reposition(n);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }

      return new logical(success);
   }

   /**
    * Conversion of REPOSITION-TO-ROWID() method (KW_REPOS_2I).
    *
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    *
    * @param   id1
    *          A primary key ID, representing the target record (in the case
    *          of a join, this ID is associated with the left-most record in
    *          the join).
    * @param   joinIDs
    *          All remaining primary key IDs, if any, arranged from left to
    *          right to coincide with the records being joined by the
    *          underlying query.
    *
    * @return  <code>true</code> if REPOSITION-TO-ROWID() succeeds,
    *          else <code>false</code>.
    */
   public logical queryRepositionByID(rowid id1, rowid...joinIDs)
   {
      boolean success = true;

      try
      {
         repositionByID(id1, joinIDs);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }

      return new logical(success);
   }

   /**
    * Conversion of NUM-RESULTS attribute (KW_NUM_RES).
    *
    * Getter of NUM-RESULTS attribute
    *
    */
   public integer getNumResults()
   {
      if (prepareString != null && closed)
      {
         return new integer();
      }

      try
      {
         return getDelegate().size();
      }
      catch (UnsupportedOperationException e)
      {
         return new integer(0);
      }
   }

   /**
    * Adds one new buffer to a query object or dynamic ProDataSet object,
    * without affecting the other buffers, if any. Use the SET-BUFFERS( )
    * method to remove all prior buffers and set all buffers for the object
    * at the same time.
    *
    * @param   bufHandle
    *          A buffer handle to add.
    *
    * @return  true on success
    */
   public logical addBuffer(handle bufHandle)
   {
      if (!canAddBuffer() || !validBuffer(bufHandle, 1))
      {
         return new logical(false);
      }

      if (!closed)
      {
         close();
      }
      
      Buffer buf = (Buffer) bufHandle.get();
      return addBuffer(buf);
   }

   /**
    * Adds one new buffer to a query object or dynamic ProDataSet object, 
    * without affecting the other buffers, if any. Use the SET-BUFFERS( )
    * method to remove all prior buffers and set all buffers for the object 
    * at the same time.
    * 
    * @param   buf
    *          A buffer to add.
    *
    * @return  true on success
    */
   @Override
   public logical addBuffer(Buffer buf)
   {
      if (!canAddBuffer())
      {
         return new logical(false);
      }

      if (!closed)
      {
         close();
      }
      
      // reset the prepare string
      prepareString = null;

      // if the buffer already exists, is added anyway
      
      this.buffers.add(buf);

      return new logical(true);
   }

   /**
    * Sets buffers for the query. Any buffers previously added
    * or set are removed. Use the <code>addBuffer()</code> method to add one 
    * buffer to the object, without affecting the other buffers, if any.
    * This method is the equivalent of <code>SET-BUFFERS()</code> method 
    * from 4GL.
    *
    * @param   hBuffers
    *          Variable argument list of buffer handles to be set.
    *          In 4GL, the maximum number of buffers per query is 18.
    * 
    * @return  true on success
    */
   @Override
   public logical setBuffers(handle... hBuffers)
   {
      if (!canAddBuffer())
      {
         return new logical(false);
      }

      // validate all handles upfront
      for (int i = 0; i < hBuffers.length; i++)
      {
         if (!validBuffer(hBuffers[i], i + 1))
         {
            return new logical(false);
         }
      }
      
      if (!closed)
      {
         close();
      }
      
      this.buffers.clear();

      for (handle h : hBuffers)
      {
         addBuffer((Buffer) h.get());
      }
      
      return new logical(true);
   }

   /**
    * Sets buffers for the query. Any buffers previously added
    * or set are removed. Use the <code>addBuffer()</code> method to add one 
    * buffer to the object, without affecting the other buffers, if any.
    * This method is the equivalent of <code>SET-BUFFERS()</code> method 
    * from 4GL.
    * 
    * @param   bufs
    *          Variable argument list of Buffer to be set.
    *          In 4GL, the maximum number of buffers per query is 18.
    * 
    * @return  true on success
    */
   @Override
   public logical setBuffers(Buffer... bufs)
   {
      if (!canAddBuffer())
      {
         return new logical(false);
      }
      
      if (!closed)
      {
         close();
      }
      
      this.buffers.clear();

      for (Buffer buf : bufs)
      {
         addBuffer(buf);
      }
      
      return new logical(true);
   }

   /**
    * Compiles a predicate (query condition). The query must be open after this call.
    * This method is the equivalent of <code>QUERY-PREPARE()</code> method from 4GL.
    *  
    * @param   predicate
    *          The predicate to be prepared by this query. The syntax of the string is the same
    *          of the <code>OPEN QUERY</code> statement of the 4GL.
    * 
    * @return  true on success
    */
   @Override
   public logical prepare(String predicate)
   {
      return prepare(new character(predicate));
   }

   /**
    * Compiles a predicate (query condition). The query must be open after this call.
    * This method is the equivalent of <code>QUERY-PREPARE()</code> method from 4GL.
    *  
    * @param   predicate
    *          The predicate to be prepared by this query. The syntax of the string is the same
    *          of the <code>OPEN QUERY</code> statement of the 4GL.
    * 
    * @return  true on success
    */
   @Override
   public logical prepare(character predicate)
   {
      // validate first
      if (predicate.isUnknown())
      {
         final String errMsg = 
            "QUERY-PREPARE requires a valid non-null character argument for query %s";
         
         ErrorManager.recordOrShowError(7321, String.format(errMsg, qname),
                                        false, false, false);
         
         return new logical(false);
      }
      if (buffers.isEmpty())
      {
         final String errMsg = 
            "QUERY-PREPARE requires that SET-BUFFERS has been used to assign buffers to query %s";
         
         ErrorManager.recordOrShowError(7322, String.format(errMsg, qname),
                                        false, false, false);
         
         return new logical(false);
      }

      // close the query, if opened
      close();
      
      // reset vars
      prepareString = null;
      
      // parse the predicate and instantiate a new delegate
      String qprepare = predicate.toStringMessage();
      P2JQuery query = DynamicQueryHelper.parseQuery(buffers, qprepare, qname);

      if (query != null)
      {
         this.prepareString = qprepare;
         this.delegate = query;
      }

      return new logical(query != null);
   }

   /**
    * Opens a query object.
    * This method is the equivalent of <code>QUERY-OPEN()</code> method from 4GL.
    * <u>Note</u>: You must perform <code>prepare()</code> on a query 
    * object before you perform <code>queryOpen<code> on it. 
    * 
    * @return  true on success
    */
   @Override
   public logical queryOpen()
   {
      if (prepareString == null)
      {
         final String errMsg = "QUERY-OPEN for query %s requires a previous QUERY-PREPARE";
         
         ErrorManager.recordOrShowError(7312, String.format(errMsg, qname), false, false, false);
         
         return new logical(false);
      }
      
      open();
      return new logical(true);
   }

   /**
    * Conversion of INDEX-INFORMATION attribute (KW_IDX_INFO).
    *
    * Getter of INDEX-INFORMATION attribute
    *
    * @return A character string consisting of a comma-separated list of the index or
    *         indexes the query uses at the level of join specified.
    */
   public character indexInformation()
   {
      // TODO: implement me
      return null;
   }

   /**
    * Conversion of INDEX-INFORMATION attribute (KW_IDX_INFO).
    *
    * Getter of INDEX-INFORMATION attribute
    *
    * @param n
    * An integer expression that evaluates to the level of join for which you want
    * index information.

    * @return A character string consisting of a comma-separated list of the index or
    *         indexes the query uses at the level of join specified.
    */
   public character indexInformation(int n)
   {
      // TODO: implement me
      return new character();
   }

   /**
    * Conversion of INDEX-INFORMATION attribute (KW_IDX_INFO).
    *
    * Getter of INDEX-INFORMATION attribute
    *
    * @param    n
    *           An integer expression that evaluates to the level of join for which you want index
    *           information.
    *
    * @return   A character string consisting of a comma-separated list of the index or
    *           indexes the query uses at the level of join specified.
    */
   public character indexInformation(NumberType n)
   {
      // TODO: implement me
      return new character();
   }

   /**
    * Conversion of FORWARD-ONLY attribute (KW_FWD_ONLY).
    *
    * Getter of FORWARD-ONLY attribute.
    *
    * @return current value of FORWARD-ONLY attribute.
    */
   public logical forwardOnly()
   {
      return new logical(forwardOnly);
   }

   /**
    * Conversion of FORWARD-ONLY attribute (KW_FWD_ONLY).
    *
    * Setter of FORWARD-ONLY attribute.
    *
    * @param  forwardOnly
    *         New value of FORWARD-ONLY attribute.
    */
   public void changeForwardOnly(logical forwardOnly)
   {
      if (forwardOnly.isUnknown())
      {
         final String errMsg =
            "Unable to assign UNKNOWN value to attribute FORWARD-ONLY on QUERY widget";
         
         ErrorManager.recordOrShowError(4083, errMsg, false, true, false);
         
         return;
      }

      if (!closed)
      {
         String msg1 = 
            "Forward only status may not be changed while query %s is open or being browsed";
         msg1 = String.format(msg1, qname);
         final String msg2 = "Unable to set attribute FORWARD-ONLY in widget  of type QUERY";
         
         ErrorManager.recordOrShowError(new int[] { 12354, 3131 },
                                        new String[] { msg1, msg2 }, 
                                        false, false, true);
         
         return;
      }
      
      this.forwardOnly = forwardOnly.booleanValue();
   }

   /**
    * Conversion of FORWARD-ONLY attribute (KW_FWD_ONLY).
    *
    * Setter of FORWARD-ONLY attribute.
    *
    * @param  forwardOnly
    *         New value of FORWARD-ONLY attribute.
    */
   public void changeForwardOnly(boolean forwardOnly)
   {
      changeForwardOnly(new logical(forwardOnly));
   }

   /**
    * Conversion of PREPARE_STRING attribute (KW_PREP_STR).
    *
    * Getter of PREPARE_STRING attribute.
    *
    * @return current value of PREPARE_STRING attribute.
    */
   public character prepareString()
   {
      return new character(prepareString);
   }

   /**
    * Conversion of IS-OPEN attribute (KW_IS_OPEN).
    *
    * Getter of IS-OPEN attribute.
    *
    * @return current value of IS-OPEN attribute.
    */
   public logical isOpen()
   {
      return new logical(!closed);
   }

   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      return new character(admData);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(String value)
   {
      this.admData.assign(value);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      this.admData.assign(value);
   }

   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself.
    * 
    * @return   The integer number of the children.
    */
   @Override
   public integer getUniqueID()
   {
      return new integer(uniqueID);
   }

   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      if (!dynamic)
      {
         // static queries can't be deleted.
         
         // although a static query handle obtained via HANDLE attribute can be deleted and a 
         // subsequent access of the HANDLE attribute on the static query returns a handle with a
         // different ID, P2J doesn't allow static queries to be deleted (at least until we need
         // to implement this too).
         return false;
      }
      
      cleanup();
      
      // all queries can be deleted
      this.valid = false;

      if (delegate != null && delegate != DEFAULT_DELEGATE)
      {
         // the delegate can be changed for static queries too, so force it to be removed
         DynamicQueryHelper.unregisterQuery(delegate);
      }
      
      return true;
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    */
   public logical getNext()
   {
      if (!canGet())
      {
         return new logical(false);
      }

      advanced = true;

      return getDelegate().getNext();
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lockType
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    */
   public logical getNext(LockType lockType)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      advanced = true;

      return getDelegate().getNext(lockType);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getNext(long lock)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      advanced = true;

      return getDelegate().getNext(lock);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getNext(NumberType lock)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      advanced = true;

      return getDelegate().getNext(lock);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getNext(long lock, long nowait)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      advanced = true;

      return getDelegate().getNext(lock, nowait);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getNext(long lock, NumberType nowait)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      advanced = true;

      return getDelegate().getNext(lock, nowait);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getNext(NumberType lock, long nowait)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      advanced = true;

      return getDelegate().getNext(lock, nowait);
   }

   /**
    * Conversion of GET-NEXT method (KW_GET_NEXT).
    *
    * Moves a query object's result list pointer ahead one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the next record in the query is found. If the query is not
    *          open or the next record cannot be found (query is empty or the query result list
    *          pointer is on the last row), the method returns <code>false</code>. If the query
    *          object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getNext(NumberType lock, NumberType nowait)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      advanced = true;

      return getDelegate().getNext(lock, nowait);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    */
   public logical getPrevious()
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getPrevious();
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getPrevious(long lock)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getPrevious(lock);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lockType
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getPrevious(LockType lockType)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getPrevious(lockType);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if thgetPe previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getPrevious(NumberType lock)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getPrevious(lock);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getPrevious(long lock, long nowait)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getPrevious(lock, nowait);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getPrevious(NumberType lock, long nowait)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getPrevious(lock, nowait);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getPrevious(long lock, NumberType nowait)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getPrevious(lock, nowait);
   }

   /**
    * Conversion of GET-PREV method (KW_GET_PREV).
    *
    * Moves a query object's result list pointer back one row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the previous record in the query is found. If the query is
    *          not open or the previous record cannot be found (query is empty or the query
    *          result list pointer is on the first row), the method returns <code>false</code>.
    *          If the query object handle is invalid, the method returns the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getPrevious(NumberType lock, NumberType nowait)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getPrevious(lock, nowait);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    */
   public logical getFirst()
   {
      if (!canGet() || !canGetFirst())
      {
         return new logical(false);
      }

      advanced = true;
      
      return getDelegate().getFirst();
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getFirst(long lock)
   {
      if (!canGet() || !canGetFirst())
      {
         return new logical(false);
      }

      advanced = true;
      
      return getDelegate().getFirst(lock);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getFirst(NumberType lock)
   {
      if (!canGet() || !canGetFirst())
      {
         return new logical(false);
      }

      advanced = true;
      
      return getDelegate().getFirst(lock);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lockType
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getFirst(LockType lockType)
   {
      if (!canGet() || !canGetFirst())
      {
         return new logical(false);
      }

      advanced = true;
      
      return getDelegate().getFirst(lockType);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getFirst(long lock, long nowait)
   {
      if (!canGet() || !canGetFirst())
      {
         return new logical(false);
      }

      advanced = true;
      
      return getDelegate().getFirst(lock, nowait);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getFirst(NumberType lock, long nowait)
   {
      if (!canGet() || !canGetFirst())
      {
         return new logical(false);
      }

      advanced = true;
      
      return getDelegate().getFirst(lock, nowait);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getFirst(long lock, NumberType nowait)
   {
      if (!canGet() || !canGetFirst())
      {
         return new logical(false);
      }

      advanced = true;
      
      return getDelegate().getFirst(lock, nowait);
   }

   /**
    * Conversion of GET-FIRST method (KW_GET_1ST).
    *
    * Moves a query object's result list pointer to the first row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the first record in the query is found. If the query is not
    *          open or the first record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getFirst(NumberType lock, NumberType nowait)
   {
      if (!canGet() || !canGetFirst())
      {
         return new logical(false);
      }

      advanced = true;
      
      return getDelegate().getFirst(lock, nowait);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    */
   public logical getLast()
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getLast();
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getLast(long lock)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getLast(lock);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getLast(NumberType lock)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getLast(lock);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lockType
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> has not valid internal value.
    */
   public logical getLast(LockType lockType)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getLast(lockType);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getLast(long lock, long nowait)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getLast(lock, nowait);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getLast(NumberType lock, long nowait)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getLast(lock, nowait);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getLast(long lock, NumberType nowait)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getLast(lock, nowait);
   }

   /**
    * Conversion of GET-LAST method (KW_GET_LAST).
    *
    * Moves a query object's result list pointer to the last row.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if the last record in the query is found. If the query is not
    *          open or the last record cannot be found (query is empty), the method returns
    *          <code>false</code>. If the query object handle is invalid, the method returns
    *          the Unknown value (?).
    *
    * @throws  IllegalArgumentException
    *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
    */
   public logical getLast(NumberType lock, NumberType nowait)
   {
      if (!canGet() || !canLastOrPrev())
      {
         return new logical(false);
      }

      return getDelegate().getLast(lock, nowait);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    */
   public logical getCurrent()
   {
      if (!canGet())
      {
         return new logical(false);
      }

      return getDelegate().getCurrent();
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    */
   public logical getCurrent(long lock)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      return getDelegate().getCurrent(lock);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    */
   public logical getCurrent(NumberType lock)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      return getDelegate().getCurrent(lock);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lockType of LockType
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    */
   public logical getCurrent(LockType lockType)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      return getDelegate().getCurrent(lockType);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    */
   public logical getCurrent(long lock, long nowait)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      return getDelegate().getCurrent(lock, nowait);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    */
   public logical getCurrent(NumberType lock, long nowait)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      return getDelegate().getCurrent(lock, nowait);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    */
   public logical getCurrent(long lock, NumberType nowait)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      return getDelegate().getCurrent(lock, nowait);
   }

   /**
    * Conversion of GET-CURRENT() method (KW_GET_CUR).
    *
    * Refetches the current record or records associated with the query.
    *
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if GET-CURRENT() succeeds, else <code>false</code>.
    */
   public logical getCurrent(NumberType lock, NumberType nowait)
   {
      if (!canGet())
      {
         return new logical(false);
      }

      return getDelegate().getCurrent(lock, nowait);
   }

   /**
    * This will creates an entry in the result list for the current row.  This implements the
    * 4GL CREATE-RESULT-LIST-ENTRY() method.
    * 
    * @return  true on success.
    */
   @Override
   public logical createResultListEntry()
   {
      boolean success = true;
      try
      {
         // TODO: implement this method ..
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * This will delete the current row in the query's results.  This implements the
    * 4GL DELETE-RESULT-LIST-ENTRY() method.
    * 
    * @return  true on success.
    */
   @Override
   public logical deleteResultListEntry()
   {
      boolean success = true;
      try
      {
         // TODO: implement this method ..
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Conversion of NUM-BUFFERS() method (KW_NUM_BUFF).
    *
    * @return  the number of buffers in a query or ProDataSet object.
    */
   public integer numBuffers()
   {
      return dynamic || delegate == DEFAULT_DELEGATE ? new integer(buffers.size()) 
                                                     : getDelegate().numBuffers();
   }

   /**
    * Specifies the fields included in a record retrieval.
    *
    * @param dmo
    *        DMO proxy which defines the buffer for which the fields are specified.
    * @param fields
    *        The fields included in a record retrieval. If no fields are specified, only record
    *        identifiers are retrieved. No subscripts can be specified for extent fields.
    */
   public void include(DataModelObject dmo, String ... fields)
   {
      getDelegate().include(dmo, fields);
   }

   /**
    * Specifies the fields excluded from a record retrieval.
    *
    * @param dmo
    *        DMO proxy which defines the buffer for which the fields are specified.
    * @param fields
    *        The fields excluded from a record retrieval. If no fields are specified, complete
    *        records are retrieved. No subscripts can be specified for extent fields.
    */
   public void exclude(DataModelObject dmo, String ... fields)
   {
      getDelegate().exclude(dmo, fields);
   }

   /**
    * Add the buffers associated with the given query to this QUERY's buffer list.
    * <p>
    * Will be a no-op when this is a {@link #dynamic} query. 
    * 
    * @param    query
    *           The P2J query from which the buffers are loaded.
    * @param    clear
    *           When <code>true</code>, it will clear the {@link #buffers} list and the predicate.
    */
   private void addBuffers(P2JQuery query, boolean clear)
   {
      if (dynamic)
      {
         return;
      }

      if (clear)
      {
         this.prepareString = null;
         buffers.clear();
      }
      
      if (query instanceof Joinable)
      {
         for (RecordBuffer buf : ((Joinable) query).getRecordBuffers())
         {
            this.buffers.add((Buffer) buf.getDMOProxy());
         }
      }
   }
   
   /**
    * Check if the resource referenced by given handle is a valid {@link Buffer}.
    * 
    * @param    h
    *           The handle to be checked.
    * @param    idx
    *           The index of this handle in a {@link #setBuffers} call.
    *           
    * @return   <code>true</code> if the handle is valid and the referrent is a {@link Buffer}.
    */
   private boolean validBuffer(handle h, int idx)
   {
      boolean valid = h._isValid() && h.getResource() instanceof Buffer;
      
      if (!valid)
      {
         final String errMsg = "ADD/SET-BUFFERS argument %d was invalid or not found";
         
         ErrorManager.recordOrShowError(7319, String.format(errMsg, idx), false, false, false);
      }
      
      return valid;
   }
   
   /**
    * Check if a buffer can be added to this query.
    * 
    * @return   <code>true</code> if a buffer can be added.
    */
   private boolean canAddBuffer()
   {
      if (!dynamic)
      {
         final String errMsg = "ADD/SET-BUFFERS may not be used on static query %s";
         
         ErrorManager.recordOrShowError(7317, String.format(errMsg, qname),
                                        false, false, false);
      }
      
      return dynamic;
   }
   
   /**
    * Check if the query is opened. If is not open, an error message is displayed (but no error
    * recorded).
    * 
    * @return   <code>true</code> if the query is not closed.
    */
   private boolean canGet()
   {
      if (closed)
      {
         final String errMsg = "Cannot run GET methods on query %s until it is opened";
         
         ErrorManager.recordOrShowError(7313, String.format(errMsg, qname),
                                        false, false, false);
      }
      
      return !closed;
   }

   /**
    * Check if the query can be advanced via GET-FIRST.
    * 
    * @return   <code>true</code> if the query can be advanced via GET-FIRST.
    */
   private boolean canGetFirst()
   {
      if (forwardOnly && advanced)
      {
         final String errMsg =
            "Cannot move Query %s back to FIRST when FORWARD-ONLY or -noautoreslist enabled";
         
         ErrorManager.recordOrShowError(12394, String.format(errMsg, qname),
                                        false, false, false);
         
         return false;
      }
      
      return true;
   }

   /**
    * Check if the query is opened. If is not open, an error message is displayed (but no error
    * recorded).
    * 
    * @return   <code>true</code> if the query is not closed.
    */
   private boolean canGetStmt()
   {
      if (closed)
      {
         final String errMsg = "Cannot Get on query %s which is not opened";
         
         ErrorManager.recordOrShowError(3159, String.format(errMsg, qname),
                                        false, false, true);
      }
      
      return !closed;
   }

   /**
    * Check if the GET-LAST or GET-FIRST can be executed.
    * 
    * @return   <code>true</code> if not in forward-only mode.
    */
   private boolean canLastOrPrev()
   {
      if (forwardOnly)
      {
         final String errMsg =
          "Cannot LAST or PREV Query %s when FORWARD-ONLY or -noautoreslist  or BREAK BY enabled";
         
         ErrorManager.recordOrShowError(12393, String.format(errMsg, qname),
                                        false, false, false);
      }
      
      return !forwardOnly;
   }
   
   /**
    * Assign a <code>P2JQuery</code> instance as this container's active
    * delegate query.  Set any deferred state which has been set for the
    * wrapper, such as scrollability and whether delegate should error out on
    * failed first/last/unique retrievals.
    *
    * @param   delegate
    *          New delegate query.
    */
   private void assignImpl(P2JQuery delegate)
   {
      if (this.closed == false && this.delegate != null)
      {
         close();
      }
      
      this.delegate = delegate;
      
      if (scrolling)
      {
         delegate.setScrolling();
      }
      
      delegate.setStandalone(false);
      delegate.setBrowsed(browsed);
      delegate.setErrorIfNull(errorIfNull);
      delegate.setFetchOnReposition(fetchOnReposition);
      
      Iterator<RepositionListener> iter = repoListeners.iterator();
      
      while (iter.hasNext())
      {
         RepositionListener next = iter.next();
         delegate.addRepositionListener(next);
      }
      
      this.closed = false;
   }

   /**
    * Convenience method to access the delegate query as an instance of
    * <code>CompoundQuery</code>.
    *
    * @return  Delegate query, cast to a <code>CompoundQuery</code>.
    *
    * @throws  IllegalStateException
    *          if the query delegate is not an instance of
    *          <code>CompoundQuery</code>.
    */
   private CompoundQuery compoundDelegate()
   {
      try
      {
         return (CompoundQuery) delegate;
      }
      catch (ClassCastException exc)
      {
         throw new IllegalStateException(
            "Expected compound query delegate;  found " +
            Utils.getUnqualifiedName(delegate.getClass()));
      }
   }
   
   /**
    * Convenience method to access the delegate query as an instance of
    * <code>PreselectQuery</code>.
    *
    * @return  Delegate query, cast to a <code>PreselectQuery</code>.
    *
    * @throws  IllegalStateException
    *          if the query delegate is not an instance of
    *          <code>PreselectQuery</code>.
    */
   private PreselectQuery preselectDelegate()
   {
      try
      {
         return (PreselectQuery) delegate;
      }
      catch (ClassCastException exc)
      {
         throw new IllegalStateException(
            "Expected preselect query delegate;  found " +
            Utils.getUnqualifiedName(delegate.getClass()));
      }
   }
   
   /**
    * Convenience method to access the delegate query as an instance of
    * <code>PresortDelegate</code>.
    *
    * @return  Delegate query, cast to a <code>PresortDelegate</code>.
    *
    * @throws  IllegalStateException
    *          if the query delegate is not an instance of
    *          <code>PresortDelegate</code>.
    */
   private PresortDelegate presortDelegate()
   {
      try
      {
         return (PresortDelegate) delegate;
      }
      catch (ClassCastException exc)
      {
         throw new IllegalStateException(
            "Expected presort query delegate;  found " +
            Utils.getUnqualifiedName(delegate.getClass()));
      }
   }
   
   /**
    * Convenience method to access the delegate query as an instance of
    * <code>RandomAccessQuery</code>.
    *
    * @return  Delegate query, cast to a <code>RandomAccessQuery</code>.
    *
    * @throws  IllegalStateException
    *          if the query delegate is not an instance of
    *          <code>RandomAccessQuery</code>.
    */
   private RandomAccessQuery randomAccessDelegate()
   {
      try
      {
         return (RandomAccessQuery) delegate;
      }
      catch (ClassCastException exc)
      {
         throw new IllegalStateException(
            "Expected random access query delegate;  found " +
            Utils.getUnqualifiedName(delegate.getClass()));
      }
   }
   
   /**
    * Convenience method to access the current delegate query.
    *
    * @return  Delegate query assigned to this wrapper by business logic, or
    *          the default delegate query.
    */
   private P2JQuery getDelegate()
   {
      P2JQuery query = (closing ? DEFAULT_DELEGATE : delegate);
      
      return query;
   }
   
   /**
    * A no-op implementation of <code>P2JQuery</code> which is used as the
    * default delegate for the enclosing <code>QueryWrapper</code>.  This
    * implementation behaves as a query which returns no results would,
    * except that its <code>isValid</code> method returns <code>false</code>,
    * as an indication that it is a "fake" query for those callers to whom
    * this information is relevant.
    */
   private static class DefaultDelegate
   implements P2JQuery
   {
      /**
       * No-op.
       */
      public void setScrolling()
      {
      }
      
      /**
       * No-op.
       */
      public void open()
      {
      }
      
      /**
       * No-op.
       */
      public void close()
      {
      }
      
      /**
       * Get the number of tables joined by this query.
       * 
       * @return  0.
       */
      public int getTableCount()
      {
         return 0;
      }
      
      /**
       * Get all the off-end listeners associated with this query.
       * 
       * @return  the off-end listeners associated with this query.
       */
      public List<QueryOffEndListener> getOffEndListeners()
      {
         return Collections.EMPTY_LIST;
      }

     /**
       * No-op.
       *
       * @param   browsed
       *          Not used.
       */
      public void setBrowsed(boolean browsed)
      {
      }
      
      /**
       * No-op.
       *
       * @param   errorIfNull
       *          Not used.
       */
      public void setErrorIfNull(boolean errorIfNull)
      {
      }
      
      /**
       * No-op.
       *
       * @param   accum
       *          Not used.
       */
      public void addAccumulator(Accumulator accum)
      {
      }
      
      /**
       * No-op.
       *
       * @param   accum
       *          Not used.
       * @param   deferred
       *          Not used.
       */
      public void addAccumulator(Accumulator accum, boolean deferred)
      {
      }
      
      /**
       * No-op.
       */
      public void first()
      {
      }
      
      /**
       * No-op.
       *
       * @param   lockType
       *          Not used.
       */
      public void first(LockType lockType)
      {
      }
      
      /**
       * No-op.
       */
      public void last()
      {
      }
      
      /**
       * No-op.
       *
       * @param   lockType
       *          Not used.
       */
      public void last(LockType lockType)
      {
      }
      
      /**
       * No-op.
       */
      public void next()
      {
      }
      
      /**
       * No-op.
       *
       * @param   lockType
       *          Not used.
       */
      public void next(LockType lockType)
      {
      }
      
      /**
       * No-op.
       */
      public void previous()
      {
      }
      
      /**
       * No-op.
       *
       * @param   lockType
       *          Not used.
       */
      public void previous(LockType lockType)
      {
      }
      
      /**
       * No-op.
       */
      public void unique()
      {
      }
      
      /**
       * No-op.
       *
       * @param   lockType
       *          Not used.
       */
      public void unique(LockType lockType)
      {
      }
      
      /**
       * No-op.
       */
      public void current()
      {
      }
      
      /**
       * No-op.
       *
       * @param   lockType
       *          Not used.
       */
      public void current(LockType lockType)
      {
      }
      
      /**
       * Default implementation which reports query is not set to fetch the
       * next record after a reposition.
       *
       * @return  <code>false</code>.
       */
      public boolean isFetchOnReposition()
      {
         return false;
      }
      
      /**
       * No-op.
       *
       * @param   enable
       *          Not used.
       */
      public void setFetchOnReposition(boolean enable)
      {
      }
      
      /**
       * No-op.
       *
       * @param   listener
       *          Not used.
       */
      public void addRepositionListener(RepositionListener listener)
      {
      }
      
      /**
       * No-op.
       *
       * @param   id
       *          Not used.
       */
      public void repositionByID(recid id)
      {
      }
      
      /**
       * No-op.
       *
       * @param   id1
       *          Not used.
       * @param   joinIDs
       *          Not used.
       */
      public void repositionByID(rowid id1, rowid... joinIDs)
      {
      }
      
      /**
       * No-op.
       *
       * @param   row
       *          Not used.
       */
      public void reposition(NumberType row)
      {
      }
      
      /**
       * No-op.
       *
       * @param   row
       *          Not used.
       */
      public void reposition(int row)
      {
      }
      
      /**
       * No-op.
       *
       * @param   rows
       *          Not used.
       */
      public void forward(NumberType rows)
      {
      }
      
      /**
       * No-op.
       *
       * @param   rows
       *          Not used.
       */
      public void forward(int rows)
      {
      }
      
      /**
       * No-op.
       *
       * @param   rows
       *          Not used.
       */
      public void backward(NumberType rows)
      {
      }
      
      /**
       * No-op.
       *
       * @param   rows
       *          Not used.
       */
      public void backward(int rows)
      {
      }
      
      /**
       * Default implementation which reports unknown value to indicate a
       * closed query.
       *
       * @return  unknown value.
       */
      public integer size()
      {
         return (new integer());
      }
      
      /**
       * Default implementation which reports unknown value as the current
       * row to indicate an empty results list.
       *
       * @return  Unknown value.
       */
      public integer currentRow()
      {
         return currentRowImpl();
      }
      
      /**
       * Default implementation which reports unknown value as the current
       * row to indicate an empty results list.
       *
       * @return  Unknown value.
       */
      public integer currentRowImpl()
      {
         return (new integer());
      }
      
      /**
       * Default implementation which reports the query is off-end.
       *
       * @return  <code>true</code>.
       */
      public logical isOffEnd()
      {
         return (new logical(true));
      }
      
      /**
       * Default implementation which reports the query is off-end.
       *
       * @return  <code>true</code>.
       */
      public boolean _isOffEnd()
      {
         return true;
      }

      /**
       * Default implementation which reports that the query has the front
       * off-end.
       *
       * @return  {@link OffEnd#FRONT} constant.
       */
      public OffEnd getOffEnd()
      {
         return OffEnd.FRONT;
      }

      /**
       * Indicate whether this query preselects all of its results.
       *
       * @return  <code>false</code>.
       */
      public boolean isPreselect()
      {
         return false;
      }

      /**
       * No-op.
       * 
       * @param   closed
       *          Not used.
       */
      public void notifyRepositionListeners(boolean closed)
      {
      }

      /**
       * Indicate whether this query stands alone, or is the component of
       * another query, or is wrapped by a query wrapper.
       * 
       * @return  <code>false</code>.
       */
      public boolean isStandalone()
      {
         return false;
      }
      
      /**
       * Change the standalone nature of this query.  A query is standalone by
       * default, but can be changed to non-standalone by the wrapper or
       * container which contains it.
       * <p>
       * This implementation is a no-op.
       * 
       * @param   standalone
       *          <code>true</code> to set query as standalone;
       *          <code>false</code> to set query as contained.
       */
      public void setStandalone(boolean standalone)
      {
      }
      
      /**
       * No-op.
       */
      public void cleanup()
      {
      }

      /**
       * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
       *
       * Returns a handle of a particular buffer of a query object.
       *
       * @param   bufferName
       *          buffer-name. A CHARACTER expression that evaluates to the name
       *          of a buffer in the query or ProDataSet object.
       *
       * @return  Unknown buffer handle.
       */
      public handle bufferHandle(String bufferName)
      {
         return (new handle());
      }

      /**
       * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
       *
       * Returns a handle of a particular buffer of a query object.
       *
       * @param   bufferName
       *          buffer-name. A CHARACTER expression that evaluates to the name
       *          of a buffer in the query or ProDataSet object.
       *
       * @return  Unknown buffer handle.
       */
      public handle bufferHandle(character bufferName)
      {
         return (new handle());
      }

      /**
       * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
       *
       * Returns a handle of a particular buffer of a query object.
       *
       * @param   bufferSequenceNumber
       *          buffer-sequence-number. An INTEGER that represents the sequence
       *          number of the desired buffer.
       *
       * @return  Unknown buffer handle.
       */
      public handle bufferHandle(int bufferSequenceNumber)
      {
         return (new handle());
      }

      /**
       * Conversion of GET-BUFFER-HANDLE() method (KW_GET_BUFH).
       *
       * Returns a handle of a particular buffer of a query object.
       *
       * @param   bufferSequenceNumber
       *          buffer-sequence-number. An INTEGER that represents the sequence
       *          number of the desired buffer.
       *
       * @return  Unknown buffer handle.
       */
      public handle bufferHandle(integer bufferSequenceNumber)
      {
         return (new handle());
      }

      /**
       * Conversion of QUERY-CLOSE() method (KW_QRY_CLOS).
       *
       * Closes a query object.
       *
       * @return  <code>false</code>.
       */
      public logical queryClose()
      {
         return (new logical(false));
      }

      /**
       * Conversion of REPOSITION-FORWARD() method (KW_REPOS_F).
       *
       * Moves a query object's result list pointer forward a particular number
       * of rows.
       *
       * @return  <code>false</code>.
       */
      public logical queryForward(int rows)
      {
         return (new logical(false));
      }

      /**
       * Conversion of REPOSITION-FORWARD() method (KW_REPOS_F).
       *
       * Moves a query object's result list pointer forward a particular number
       * of rows.
       *
       * @return  <code>false</code>.
       */
      public logical queryForward(integer rows)
      {
         return (new logical(false));
      }

      /**
       * Conversion of REPOSITION-TO-ROW() method (KW_REPOS_2R).
       *
       * Moves a query object's result list pointer to the row corresponding to the specified
       * sequence number.
       *
       * @param   n
       *          An integer expression representing the sequence number.
       *
       * @return  <code>false</code>.
       */
      public logical queryReposition(int n)
      {
         return (new logical(false));
      }

      /**
       * Conversion of REPOSITION-TO-ROW() method (KW_REPOS_2R).
       *
       * Moves a query object's result list pointer to the row corresponding to the specified
       * sequence number.
       *
       * @param   n
       *          An integer expression representing the sequence number.
       *
       * @return  <code>false</code>.
       */
      public logical queryReposition(NumberType n)
      {
         return (new logical(false));
      }

      /**
       * Conversion of REPOSITION-TO-ROWID() method (KW_REPOS_2I).
       *
       * Reposition the cursor such that a request to retrieve the next result
       * will retrieve the result which matches the specified array of primary
       * key IDs.  A request to retrieve the previous result will retrieve the
       * result immediately previous to this, if any, in the results list.
       *
       * @param   id1
       *          A primary key ID, representing the target record (in the case
       *          of a join, this ID is associated with the left-most record in
       *          the join).
       * @param   joinIDs
       *          All remaining primary key IDs, if any, arranged from left to
       *          right to coincide with the records being joined by the
       *          underlying query.
       *
       * @return  <code>false</code>.
       */
      public logical queryRepositionByID(rowid id1, rowid...joinIDs)
      {
         return (new logical(false));
      }

      /**
       * Conversion of NUM-RESULTS attribute (KW_NUM_RES).
       *
       * Getter of NUM-RESULTS attribute
       *
       * @return  Unknown value.
       */
      public integer getNumResults()
      {
         return (new integer());
      }

      /**
       * Adds one new buffer to a query object or dynamic ProDataSet object,
       * without affecting the other buffers, if any. Use the SET-BUFFERS( )
       * method to remove all prior buffers and set all buffers for the object
       * at the same time.
       *
       * @param   bufHandle
       *          A buffer handle to add.
       *
       * @return  <code>false</code>.
       */
      public logical addBuffer(handle bufHandle)
      {
         return (new logical(false));
      }

      /**
       * Adds one new buffer to a query object or dynamic ProDataSet object, 
       * without affecting the other buffers, if any. Use the SET-BUFFERS( )
       * method to remove all prior buffers and set all buffers for the object 
       * at the same time.
       * 
       * @param   buf
       *          A buffer to add.
       *
       * @return  <code>false</code>.
       */
      @Override
      public logical addBuffer(Buffer buf)
      {
         return (new logical(false));
      }

      /**
       * Sets buffers for the query. Any buffers previously added
       * or set are removed. Use the <code>addBuffer()</code> method to add one 
       * buffer to the object, without affecting the other buffers, if any.
       * This method is the equivalent of <code>SET-BUFFERS()</code> method 
       * from 4GL.
       *
       * @param   hBuffers
       *          Variable argument list of buffer handles to be set.
       *          In 4GL, the maximum number of buffers per query is 18.
       *
       * @return  <code>false</code>.
       */
      @Override
      public logical setBuffers(handle... hBuffers)
      {
         return (new logical(false));
      }

      /**
       * Sets buffers for the query. Any buffers previously added
       * or set are removed. Use the <code>addBuffer()</code> method to add one 
       * buffer to the object, without affecting the other buffers, if any.
       * This method is the equivalent of <code>SET-BUFFERS()</code> method 
       * from 4GL.
       * 
       * @param   bufs
       *          Variable argument list of Buffer to be set.
       *          In 4GL, the maximum number of buffers per query is 18.
       * 
       * @return  true on success
       */
      @Override
      public logical setBuffers(Buffer... bufs)
      {
         return (new logical(false));
      }

      /**
       * Compiles a predicate (query condition). This method is
       * the equivalent of <code>QUERY-PREPARE()</code> method from 4GL.
       * 
       * @param   predicate
       *          The predicate to be prepared by this query.
       *
       * @return  <code>false</code>.
       */
      @Override
      public logical prepare(String predicate)
      {
         return (new logical(false));
      }

      /**
       * Compiles a predicate (query condition). This method is
       * the equivalent of <code>QUERY-PREPARE()</code> method from 4GL.
       * 
       * @param   predicate
       *          The predicate to be prepared by this query.
       *
       * @return  <code>false</code>.
       */
      @Override
      public logical prepare(character predicate)
      {
         return (new logical(false));
      }

      /**
       * Opens a query object. 
       * This method is the equivalent of <code>QUERY-OPEN()</code> method from 4GL.
       * <u>Note</u>: You must perform <code>prepare()</code> on a query 
       * object before you perform <code>queryOpen<code> on it. 
       * 
       * @return  true on success
       */
      @Override
      public logical queryOpen()
      {
         return (new logical(false));
      }

      /**
       * Conversion of INDEX-INFORMATION attribute (KW_IDX_INFO).
       *
       * Getter of INDEX-INFORMATION attribute.
       *
       * @return  Unknown value.
       */
      public character indexInformation()
      {
         return (new character());
      }

      /**
       * Conversion of INDEX-INFORMATION attribute (KW_IDX_INFO).
       *
       * Getter of INDEX-INFORMATION attribute
       *
       * @param n
       * An integer expression that evaluates to the level of join for which you want
       * index information.

       * @return  Unknown value.
       */
      public character indexInformation(int n)
      {
         return (new character());
      }

      /**
       * Conversion of INDEX-INFORMATION attribute (KW_IDX_INFO).
       *
       * Getter of INDEX-INFORMATION attribute
       *
       * @param    n
       *           An integer expression that evaluates to the level of join for which you want 
       *           index information.
       *           
       * @return   Unknown value.
       */
      public character indexInformation(NumberType n)
      {
         return (new character());
      }

      /**
       * Conversion of FORWARD-ONLY attribute (KW_FWD_ONLY).
       *
       * Getter of FORWARD-ONLY attribute.
       *
       * @return  <code>false</code>.
       */
      public logical forwardOnly()
      {
         return (new logical(false));
      }

      /**
       * Conversion of FORWARD-ONLY attribute (KW_FWD_ONLY).
       *
       * Setter of FORWARD-ONLY attribute.
       *
       * @param  forwardOnly
       *         New value of FORWARD-ONLY attribute.
       */
      public void changeForwardOnly(logical forwardOnly)
      {
      }

      /**
       * Conversion of FORWARD-ONLY attribute (KW_FWD_ONLY).
       *
       * Setter of FORWARD-ONLY attribute.
       *
       * @param  forwardOnly
       *         New value of FORWARD-ONLY attribute.
       */
      public void changeForwardOnly(boolean forwardOnly)
      {
      }

      /**
       * Conversion of PREPARE_STRING attribute (KW_PREP_STR).
       *
       * Getter of PREPARE_STRING attribute.
       *
       * @return  Unknown value.
       */
      public character prepareString()
      {
         return (new character());
      }

      /**
       * Conversion of IS-OPEN attribute (KW_IS_OPEN).
       *
       * Getter of IS-OPEN attribute.
       *
       * @return current value of IS-OPEN attribute.
       */
      public logical isOpen()
      {
         return (new logical(false));
      }

      /**
       * Conversion of GET-NEXT method (KW_GET_NEXT).
       *
       * Moves a query object's result list pointer ahead one row.
       *
       * @return  <code>false</code>.
       */
      public logical getNext()
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-NEXT method (KW_GET_NEXT).
       *
       * Moves a query object's result list pointer ahead one row.
       *
       * @param   lockType
       *          Lock type.
       *
       * @return  <code>false</code>.
       */
      public logical getNext(LockType lockType)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-NEXT method (KW_GET_NEXT).
       *
       * Moves a query object's result list pointer ahead one row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> has not valid internal value.
       */
      public logical getNext(long lock)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-NEXT method (KW_GET_NEXT).
       *
       * Moves a query object's result list pointer ahead one row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> has not valid internal value.
       */
      public logical getNext(NumberType lock)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-NEXT method (KW_GET_NEXT).
       *
       * Moves a query object's result list pointer ahead one row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getNext(long lock, long nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-NEXT method (KW_GET_NEXT).
       *
       * Moves a query object's result list pointer ahead one row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getNext(long lock, NumberType nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-NEXT method (KW_GET_NEXT).
       *
       * Moves a query object's result list pointer ahead one row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getNext(NumberType lock, long nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-NEXT method (KW_GET_NEXT).
       *
       * Moves a query object's result list pointer ahead one row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getNext(NumberType lock, NumberType nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-PREV method (KW_GET_PREV).
       *
       * Moves a query object's result list pointer back one row.
       *
       * @return  <code>false</code>.
       */
      public logical getPrevious()
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-PREV method (KW_GET_PREV).
       *
       * Moves a query object's result list pointer back one row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> has not valid internal value.
       */
      public logical getPrevious(long lock)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-PREV method (KW_GET_PREV).
       *
       * Moves a query object's result list pointer back one row.
       *
       * @param   lockType
       *          Lock type.
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> has not valid internal value.
       */
      public logical getPrevious(LockType lockType)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-PREV method (KW_GET_PREV).
       *
       * Moves a query object's result list pointer back one row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> has not valid internal value.
       */
      public logical getPrevious(NumberType lock)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-PREV method (KW_GET_PREV).
       *
       * Moves a query object's result list pointer back one row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getPrevious(long lock, long nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-PREV method (KW_GET_PREV).
       *
       * Moves a query object's result list pointer back one row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getPrevious(NumberType lock, long nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-PREV method (KW_GET_PREV).
       *
       * Moves a query object's result list pointer back one row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getPrevious(long lock, NumberType nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-PREV method (KW_GET_PREV).
       *
       * Moves a query object's result list pointer back one row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getPrevious(NumberType lock, NumberType nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-FIRST method (KW_GET_1ST).
       *
       * Moves a query object's result list pointer to the first row.
       *
       * @return  <code>false</code>.
       */
      public logical getFirst()
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-FIRST method (KW_GET_1ST).
       *
       * Moves a query object's result list pointer to the first row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> has not valid internal value.
       */
      public logical getFirst(long lock)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-FIRST method (KW_GET_1ST).
       *
       * Moves a query object's result list pointer to the first row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> has not valid internal value.
       */
      public logical getFirst(NumberType lock)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-FIRST method (KW_GET_1ST).
       *
       * Moves a query object's result list pointer to the first row.
       *
       * @param   lockType
       *          Lock type.
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> has not valid internal value.
       */
      public logical getFirst(LockType lockType)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-FIRST method (KW_GET_1ST).
       *
       * Moves a query object's result list pointer to the first row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getFirst(long lock, long nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-FIRST method (KW_GET_1ST).
       *
       * Moves a query object's result list pointer to the first row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getFirst(NumberType lock, long nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-FIRST method (KW_GET_1ST).
       *
       * Moves a query object's result list pointer to the first row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getFirst(long lock, NumberType nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-FIRST method (KW_GET_1ST).
       *
       * Moves a query object's result list pointer to the first row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getFirst(NumberType lock, NumberType nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-LAST method (KW_GET_LAST).
       *
       * Moves a query object's result list pointer to the last row.
       *
       * @return  <code>false</code>.
       */
      public logical getLast()
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-LAST method (KW_GET_LAST).
       *
       * Moves a query object's result list pointer to the last row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> has not valid internal value.
       */
      public logical getLast(long lock)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-LAST method (KW_GET_LAST).
       *
       * Moves a query object's result list pointer to the last row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> has not valid internal value.
       */
      public logical getLast(NumberType lock)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-LAST method (KW_GET_LAST).
       *
       * Moves a query object's result list pointer to the last row.
       *
       * @param   lockType
       *          Lock type.
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> has not valid internal value.
       */
      public logical getLast(LockType lockType)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-LAST method (KW_GET_LAST).
       *
       * Moves a query object's result list pointer to the last row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getLast(long lock, long nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-LAST method (KW_GET_LAST).
       *
       * Moves a query object's result list pointer to the last row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getLast(NumberType lock, long nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-LAST method (KW_GET_LAST).
       *
       * Moves a query object's result list pointer to the last row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getLast(long lock, NumberType nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-LAST method (KW_GET_LAST).
       *
       * Moves a query object's result list pointer to the last row.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       *
       * @throws  IllegalArgumentException
       *          if <code>lock</code> or <code>nowait</code> have not valid internal values.
       */
      public logical getLast(NumberType lock, NumberType nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @return  <code>false</code>.
       */
      public logical getCurrent()
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       *
       * @return  <code>false</code>.
       */
      public logical getCurrent(long lock)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       *
       * @return  <code>false</code>.
       */
      public logical getCurrent(NumberType lock)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lockType
       *          Lock type.
       *
       * @return  <code>false</code>.
       */
      public logical getCurrent(LockType lockType)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       */
      public logical getCurrent(long lock, long nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       */
      public logical getCurrent(NumberType lock, long nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       */
      public logical getCurrent(long lock, NumberType nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of GET-CURRENT() method (KW_GET_CUR).
       *
       * Refetches the current record or records associated with the query.
       *
       * @param   lock
       *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
       * @param   nowait
       *          NO-WAIT
       *
       * @return  <code>false</code>.
       */
      public logical getCurrent(NumberType lock, NumberType nowait)
      {
         return new logical(false);
      }

      /**
       * Conversion of NUM-BUFFERS() method (KW_NUM_BUFF).
       *
       * @return  the number of buffers in a query or ProDataSet object.
       */
      public integer numBuffers()
      {
         return new integer();
      }

      /**
       * No-op.
       *
       * @param dmo
       *        Not used.
       * @param fields
       *        Not used.
       */
      public void include(DataModelObject dmo, String ... fields)
      {
      }

      /**
       * No-op.
       *
       * @param dmo
       *        Not used.
       * @param fields
       *        Not used.
       */
      public void exclude(DataModelObject dmo, String ... fields)
      {
      }
   }
}
