/*
** Module   : TempLock.java
** Abstract : An object which encapsulates a resettable lock on a specific DMO
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20061112   @31143 Created initial version. An object which encapsulates a resettable
**                           lock on a specific DMO.
** 002 ECF 20070412   @32991 Retrofit to use RecordLockContext for record locking. Minor
**                           implementation change.
** 003 ECF 20131029          Import change. Removed unused instance variable.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import com.goldencode.p2j.persist.lock.*;

/**
 * Object which encapsulates a resettable lock on a specific DMO.  Successful
 * construction of an instance of this class indicates that the requested
 * lock type was obtained from the appropriate lock manager or was already
 * available to this session.  The old lock type is restored by invoking
 * {@link #reset}.
 */
public final class TempLock
{
   /** Unique record identifier */
   private final RecordIdentifier ident;
   
   /** Previous lock type of DMO */
   private final LockType originalLockType;
   
   /** Object which services lock requests */
   private final Persistence persistence;
   
   /**
    * Constructor which acquires a record lock of the specified type.  This
    * constructor may block indefinitely if the lock is unavailable and the
    * requested lock type is not a NO_WAIT variant.
    *
    * @param   persistence
    *          Object which services lock requests.
    * @param   lockType
    *          Type of lock to be acquired.
    * @param   table
    *          Table containing target DMO.
    * @param   recordID
    *          Primary key ID of DMO being locked.
    */
   TempLock(Persistence persistence, LockType lockType, String table, Serializable recordID)
   throws LockUnavailableException
   {
      this.persistence = persistence;
      this.ident = new RecordIdentifier(table, recordID);
      
      LockType currentType = persistence.queryLock(ident);
      
      if (lockType.compareTo(currentType) > 0)
      {
         originalLockType = currentType;
         RecordLockContext lockCtx = persistence.getRecordLockContext();
         lockCtx.lock(ident, lockType);
      }
      else
      {
         originalLockType = null;
      }
   }
   
   /**
    * Reset the record lock back to its original type.  This is a no-op if
    * the old and new types are the same or if the lock type was never
    * changed by this object.
    */
   void reset()
   {
      if (originalLockType == null)
      {
         // Lock type was never changed;  nothing to do.
         return;
      }
      
      LockType currentType = persistence.queryLock(ident);
      
      if (currentType.compareTo(originalLockType) > 0)
      {
         try
         {
            RecordLockContext lockCtx = persistence.getRecordLockContext();
            lockCtx.lock(ident, originalLockType);
         }
         catch (LockUnavailableException exc)
         {
            // Not thrown when requesting non-NO_WAIT lock type variants.
         }
      }
   }
}
