/*
** Module   : DefaultDirtyShareContext.java
** Abstract : Default implementation of DirtyShareContext
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20080420   @38626 Created initial version. Client context's
**                           primary bridge to the DirtyShareManager.
**                           Manages dirty state specific to the current
**                           context.
** 002 ECF 20080610   @38682 Added safety logic to prevent NPEs.
** 003 ECF 20080612   @38719 Fixed getDirtyInfo(). If an uncommitted
**                           delete was made before the backing table in
**                           the dirty database was created, we could
**                           attempt a query nonetheless and bomb.
** 004 ECF 20080627   @39073 Minor optimization. Avoid collection empty
**                           check is some cases.
** 005 CA  20080716   @39122 Renamed method lockUniqueIndexes to 
**                           lockAllIndexes - this will lock all the 
**                           indexes for a certain entity. Modified the
**                           delete(...) and rollbackInsert(...) methods
**                           so the index locking can be disabled.
** 006 ECF 20080815   @39542 Added support for global change events. These
**                           are events that represent index-changing
**                           inserts, updates, or deletes, which are
**                           processed by sessions other than the session
**                           in which the event originated.
** 007 ECF 20081211   @40869 Fixed delete()/rollbackInsert() defect. The
**                           deletion of a record which was inserted in
**                           the same transaction rolls back the insert,
**                           but does not track the record as deleted.
**                           This caused a problem when the insert itself
**                           was later rolled back. Such records are now
**                           tracked via the deadInserts set.
** 008 ECF 20090428   @42129 Added isLocalInsert(). Reports whether a record
**                           is being tracked as newly inserted by this
**                           context.
** 009 ECF 20090513   @42173 Modified update() method. Need to pass new lock
**                           parameter to modified DirtyShareManager update()
**                           method.
** 010 ECF 20090601   @42569 Added support for Progress isolation leak quirk.
**                           Added new methods.
** 011 ECF 20090610   @42639 Modified updateSnapshots() method. Now returns
**                           list of DMOs affected and tracks entities whose
**                           uncommitted changes have been leaked. The list of
**                           such entities is now used during cleanup.
** 012 ECF 20090611   @42661 Optimized updateSnapshots(). In a long-running
**                           transaction with many updates, it was doing much
**                           unnecessary work. We now track which DMOs need to
**                           have their snapshots updated and refresh only
**                           these.
** 013 CA  20090721   @43285 Fixed rollback of dirty records, when a record is
**                           in dead insert mode - the changes map should not 
**                           be checked in this case.
** 014 ECF 20090721   @43307 Modified delete(). If the record being deleted is
**                           already being tracked for changes, it must be
**                           removed from the changes and dirtySnapshots
**                           collections.
** 015 ECF 20090819   @43692 Added new update() variant. Used when it is known
**                           that the target DMO already is being tracked by
**                           the dirty share manager.
** 016 ECF 20110405          Fixed dirty change leak. Remove record identifier
**                           from deadInserts set upon insert to avoid error
**                           during rollbackInsert() for same record later.
**                           Otherwise, a record is leaked in the dirty share
**                           manager for the life of the server process. Also
**                           fixed (separate) minor bug in rollbackInsert()
**                           which was not removing dirtySnapshot identifiers
**                           when their corresponding records were removed
**                           from dirty share tracking.
** 017 ECF 20131028          Import change.
*/

package com.goldencode.p2j.persist.dirty;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.util.*;
import org.hibernate.type.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;

/**
 * Default implementation of a client context's primary interface to store
 * uncommitted index updates for sharing with other contexts, and to read
 * uncommitted index updates made by other contexts.  This is intended to
 * support compatibility with Progress' peculiar transaction isolation
 * behavior of immediately sharing updates to indexed fields across user
 * sessions, but deferring the sharing of updates to non-indexed fields until
 * the enclosing transaction is committed.
 * <p>
 * This implementation maintains state specific to the current context, in
 * order both to filter information appropriately, and to be able to clean up
 * shared data when the current context's transactions are completed.
 * 
 * @see  DirtyShareManager
 * @see  DefaultDirtyShareManager
 * @see  NopDirtyShareContext
 */
public final class DefaultDirtyShareContext
implements DirtyShareContext
{
   /** Object which manages dirty reads for the database */
   private final DirtyShareManager manager;
   
   /** Records inserted by this context */
   private Set<RecordIdentifier> inserts = null;
   
   /** Records inserted by this context, which deletes have rolled back */
   private Set<RecordIdentifier> deadInserts = null;
   
   /** Map of entity names to sets of IDs inserted/updated by this context */
   private Map<String, Set<Serializable>> changes = null;
   
   /** Map of entity names to sets of IDs deleted by this context */
   private Map<String, Set<Serializable>> deletes = null;
   
   /** Set of IDs of DMOs changed since last updateSnapshots() call */
   private Set<RecordIdentifier> dirtySnapshots = null;
   
   /** DMO entities which have had uncommitted updates "leaked" */
   private Set<String> leakedEntities = null;
   
   /**
    * Constructor which stores a reference to the dirty share manager for the
    * associated database.
    * 
    * @param   manager
    *          Dirty share manager for the associated database.
    */
   DefaultDirtyShareContext(DirtyShareManager manager)
   {
      this.manager = manager;
   }
   
   /**
    * Register with a global event clearinghouse an interest in insert,
    * delete, and update events which affect the given DMO entity.
    * 
    * @param   entity
    *          DMO entity for which the caller is interested in tracking
    *          change.
    * 
    * @return  A unique identifier for the registrant, assigned by the event
    *          clearinghouse, used later to collect events and deregister.
    * 
    * @see     #getGlobalEvents(long, long)
    * @see     #deregisterForGlobalEvents(long)
    */
   public Long registerForGlobalEvents(String entity)
   {
      return manager.registerForGlobalEvents(entity);
   }
   
   /**
    * Deregister with a global event clearinghouse an interest in insert,
    * delete, and update events which affect a particular DMO type.
    * 
    * @param   registerID
    *          A unique identifier for the registrant being deregistered.
    * 
    * @see     #getGlobalEvents(long, long)
    * @see     #registerForGlobalEvents(String)
    */
   public void deregisterForGlobalEvents(long registerID)
   {
      manager.deregisterForGlobalEvents(registerID);
   }
   
   /**
    * Retrieve any new events that have been collected by a global event
    * clearinghouse, representing inserts, deletes, and updates made to
    * record types for which the caller previously has registered interest.
    * 
    * @param   registerID
    *          A unique identifier for an object which has registered interest
    *          in record-changing events.
    * @param   lastID
    *          A unique identifier representing the last event retrieved by
    *          the caller.  If less than 0, it is assumed no events have yet
    *          been retrieved since the caller registered interest in these
    *          events.
    * 
    * @return  Array of applicable, new events, or <code>null</code> if no
    *          such events have been collected.
    * 
    * @throws  EventRegistrationException
    *          if <code>registerID</code> is invalid or has expired.
    */
   public GlobalChangeEvent[] getGlobalEvents(long registerID, long lastID)
   throws EventRegistrationException
   {
      return manager.getGlobalEvents(registerID, lastID);
   }
   
   /**
    * Lock or unlock all indexes associated with the given DMO entity.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   lockType
    *          Type of lock to acquire, or <code>NONE</code> to release
    *          existing locks.  <code>NO_WAIT</code> variants will be treated
    *          as regular, blocking variants.
    * 
    * @throws  PersistenceException
    *          if there is an error querying JDBC metadata while retrieving
    *          index information.
    */
   public void lockAllIndexes(String entity, LockType lockType)
   throws PersistenceException
   {
      manager.lockAllIndexes(entity, lockType);
   }
   
   /**
    * Begin tracking a newly inserted record in the dirty database.
    * <p>
    * This record is removed when the adding context's current transaction is
    * committed or rolled back.
    * 
    * @param   buffer
    *          Record buffer responsible for the insert.
    * @param   dmo
    *          Record to be tracked.
    * @param   setter
    *          Optional setter method for a property which must be updated in
    *          the DMO copy before it is inserted into the dirty database.
    *          May be <code>null</code>.
    * @param   datum
    *          Updated value to apply to <code>property</code>.  May be
    *          <code>null</code>.
    * @param   lock
    *          <code>true</code> to acquire write locks to insert the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void insert(RecordBuffer buffer,
                      Persistable dmo,
                      Method setter,
                      BaseDataType datum,
                      boolean lock)
   throws PersistenceException
   {
      String entity = buffer.getEntityName();
      Serializable id = dmo.getId();
      RecordIdentifier ident = new RecordIdentifier(entity, id);
      
      if (inserts != null && inserts.contains(ident))
      {
         // This record was already inserted during this transaction.
         return;
      }
      
      if (deadInserts != null)
      {
         // Remove this record from the dead inserts set if we were tracking
         // it as such;  this will prevent any problems if we later have to
         // roll back this insert.
         deadInserts.remove(ident);
      }
      
      // Make a copy of the DMO.
      Persistable copy = (Persistable) dmo.deepCopy();
      
      // Apply pending update if necessary.
      if (setter != null)
      {
         try
         {
            setter.invoke(copy, datum);
         }
         catch (Exception exc)
         {
            throw new PersistenceException(exc);
         }
      }
      
      // Store in database.
      manager.insert(entity, copy, lock);
      
      // Track insert in local context.
      if (inserts == null)
      {
         inserts = new HashSet<RecordIdentifier>();
      }
      inserts.add(ident);
      trackChange(entity, id);
   }
   
   /**
    * Indicate whether the specified record is being tracked as newly inserted
    * by this context.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be checked.
    * 
    * @return  <code>true</code> if this record is newly inserted by this
    *          context, else <code>false</code>.
    */
   public boolean isLocalInsert(String entity, Serializable id)
   {
      if (inserts == null)
      {
         return false;
      }
      
      RecordIdentifier ident = new RecordIdentifier(entity, id);
      
      return inserts.contains(ident);
   }
   
   /*
   public boolean isLocalDelete(String entity, Serializable id)
   {
      if (deletes != null)
      {
         Set<Serializable> deleted = deletes.get(entity);
         if (deleted != null && deleted.contains(id))
         {
            return true;
         }
      }
      
      if (deadInserts != null)
      {
         RecordIdentifier ident = new RecordIdentifier(entity, id);
         if (deadInserts.contains(ident))
         {
            return true;
         }
      }
      
      return false;
   }
   */
   
   /**
    * Rollback the insert of a record previously introduced with the {@link
    * #insert(RecordBuffer, Persistable, Method, BaseDataType, boolean)}
    * method.
    * 
    * @param   ident
    *          Identifier for the target record, which encapsulates its entity
    *          name and primary key.
    * @param   lock
    *          <code>true</code> to acquire write locks;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  IllegalStateException
    *          if the record was not inserted in this context during the
    *          current transaction
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void rollbackInsert(RecordIdentifier ident, boolean lock)
   throws PersistenceException
   {
      // if the target record was already deleted from the dirty context, then
      // the changes map check is not relevant (it may be empty if only one
      // record is in the dirty context and this is a dead record)
      if ((deadInserts == null || !deadInserts.contains(ident)) && 
          (changes == null || changes.isEmpty() || 
           inserts == null || !inserts.contains(ident)))
      {
         throw new IllegalStateException(
            "No dirty entry for '" + ident + "';  cannot roll back insert");
      }
      
      try
      {
         if (deadInserts != null && deadInserts.remove(ident))
         {
            // This insert occurred within the existing transaction, and was
            // rolled back by a subsequent delete within the transaction, so
            // it already has been removed from the dirty share manager.  We
            // cannot remove it again as this will cause an error.
            
            if (deadInserts.isEmpty())
            {
               deadInserts = null;
            }
            
            return;
         }
         
         // Remove from local context insert tracking.
         inserts.remove(ident);
         
         // Remove from database.
         manager.rollbackInsert(ident, lock);
      }
      finally
      {
         // Remove from local context change tracking.
         String entity = ident.getTable();
         Serializable id = ident.getRecordID();
         Set<Serializable> ids = changes.get(entity);
         if (ids != null && ids.remove(id))
         {
            if (ids.isEmpty())
            {
               changes.remove(entity);
            }
            
            dirtySnapshots.remove(ident);
         }
      }
   }
   
   /**
    * Track an update to one or more indexed, DMO properties, as indicated by
    * the <code>properties</code> and <code>values</code> arguments.
    * <p>
    * Any record added to the dirty database as a result of this method is
    * removed when the adding context's current transaction is committed or
    * rolled back.
    * 
    * @param   buffer
    *          Record buffer responsible for the update.
    * @param   properties
    *          Array of names of those DMO properties which were updated.
    *          These should represent columns which participate in one or more
    *          indexes.
    * @param   values
    *          Updated DMO property values.  Each value in the array
    *          corresponds positionally with its matching property name in the
    *          <code>properties</code> array.
    * @param   lock
    *          <code>true</code> to acquire write locks to update the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void update(RecordBuffer buffer,
                      String[] properties,
                      BaseDataType[] values,
                      boolean lock)
   throws PersistenceException
   {
      String entity = buffer.getEntityName();
      Persistable snapshot = buffer.getSnapshot();
      Serializable id = snapshot.getId();
      
      // Determine whether the dirty share manager is already tracking changes
      // to this record from a previous insert or update.  If not, begin
      // tracking the record, then update it.  This set of calls is assumed to
      // be "atomic", since we must have an exclusive lock on the record.
      if (manager.isTracked(entity, id))
      {
         manager.update(entity, id, properties, values, lock);
      }
      else
      {
         Persistable copy = (Persistable) snapshot.deepCopy();
         manager.update(entity, copy, properties, values, lock);
      }
      
      // Track update in local context.
      trackChange(entity, id);
   }
   
   /**
    * Track an update to one or more indexed, DMO properties, as indicated by
    * the <code>properties</code> and <code>values</code> arguments.
    * <p>
    * It is assumed the record identified by <code>entity</code> and
    * <code>id</code> already is present in the dirty database due to a
    * previous update or insert event.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the target record.
    * @param   properties
    *          Array of names of those DMO properties which were updated.
    *          These should represent columns which participate in one or more
    *          indexes.
    * @param   values
    *          Updated DMO property values.  Each value in the array
    *          corresponds positionally with its matching property name in the
    *          <code>properties</code> array.
    * @param   lock
    *          <code>true</code> to acquire write locks to update the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public void update(String entity,
                      Serializable id,
                      String[] properties,
                      BaseDataType[] values,
                      boolean lock)
   throws PersistenceException
   {
      manager.update(entity, id, properties, values, lock);
   }
   
   /**
    * Overwrite the state of the records in the dirty database which match the
    * IDs of all DMOs for the given entities, which have been modified by the
    * current context.
    * <p>
    * This is necessary to support a quirk/bug in Progress which compromises
    * transaction isolation by leaking uncommitted changes (even non-indexed
    * data) under certain conditions.
    * 
    * @param   entities
    *          List of entities, which are the names of the DMO implementation
    *          classes associated with tables being tracked for uncommitted
    *          changes.  Only those records which match these entities will
    *          have their current state leaked.
    * @param   persistence
    *          Persistence service object to be used to retrieved the latest
    *          version of the target records in the current context.
    * 
    * @return  List of all DMOs whose snapshots were updated with the dirty
    *          share manager.  The list may be empty, but will not be
    *          <code>null</code>.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public List<Persistable> updateSnapshots(List<String> entities,
                                            Persistence persistence)
   throws PersistenceException
   {
      if (changes == null || dirtySnapshots == null)
      {
         // Nothing to update.
         return Collections.emptyList();
      }
      
      List<Persistable> allDMOs = null;
      
      for (String entity : entities)
      {
         Set<Serializable> changed = changes.get(entity);
         if (changed == null)
         {
            continue;
         }
         
         int size = changed.size();
         if (size > 0)
         {
            if (allDMOs == null)
            {
               allDMOs = new ArrayList<Persistable>();
            }
            
            List<Persistable> entityDMOs = new ArrayList<Persistable>(size);
            
            Iterator<Serializable> iter = changed.iterator();
            for (int i = 0; iter.hasNext(); i++)
            {
               Serializable id = iter.next();
               RecordIdentifier ident = new RecordIdentifier(entity, id);
               if (dirtySnapshots.remove(ident))
               {
                  Persistable dmo = persistence.quickLoad(ident);
                  entityDMOs.add(dmo);
                  allDMOs.add(dmo);
               }
            }
            
            size = entityDMOs.size();
            if (size > 0)
            {
                Persistable[] dmos =
                   entityDMOs.toArray(new Persistable[size]);
                
                manager.updateSnapshots(entity, dmos);
                
                if (leakedEntities == null)
                {
                   leakedEntities = new HashSet<String>();
                }
                
                leakedEntities.add(entity);
            }
        }
      }
      
      if (allDMOs == null)
      {
         allDMOs = Collections.emptyList();
      }
      
      return allDMOs;
   }
   
   /**
    * Mark a record as having been deleted in an uncommitted transaction, or
    * unmark a record which previously was so marked.
    * <p>
    * If marking a record as deleted, that was earlier inserted by the current
    * context, rollback the insert, but do not mark the record as deleted.
    * <p>
    * A record is unmarked when the adding context's current transaction is
    * committed or rolled back.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of deleted record.
    * @param   rollback
    *          If <code>false</code>, the target record is marked as deleted;
    *          if <code>true</code>, it is unmarked.
    * @param   lock
    *          <code>true</code> to acquire write locks to delete the record;
    *          <code>false</code> to assume write locks are acquired by
    *          calling code.
    * 
    * @throws  PersistenceException
    *          if a database error occurs rolling back an earlier insert.
    */
   public void delete(String entity, 
                      Serializable id, 
                      boolean rollback, 
                      boolean lock)
   throws PersistenceException
   {
      if (!rollback)
      {
         RecordIdentifier ident = new RecordIdentifier(entity, id);
         
         // Special case 1:  if we are deleting a record we previously
         // inserted in the same transaction, remove the record, but do not
         // update the shared deleted records state.
         if (inserts != null && inserts.contains(ident))
         {
            rollbackInsert(ident, lock);
            if (deadInserts == null)
            {
               deadInserts = new HashSet<RecordIdentifier>(8);
            }
            deadInserts.add(ident);
            
            return;
         }
         
         // Special case 2:  if we are deleting a record we previously
         // updated in the same transaction, remove the record, but still
         // update the shared deleted records state.
         if (changes != null)
         {
            Set<Serializable> changed = changes.get(entity);
            if (changed != null && changed.remove(id))
            {
               if (changed.isEmpty())
               {
                  changes.remove(entity);
               }
               dirtySnapshots.remove(ident);
            }
         }
      }
      
      // Update shared state.
      manager.delete(entity, id, rollback, lock);
      
      // Track state locally, so we can clean up at end of transaction.
      if (rollback)
      {
         if (deletes != null)
         {
            Set<Serializable> ids = deletes.get(entity);
            if (ids != null)
            {
               if (ids.remove(id) && ids.isEmpty())
               {
                  deletes.remove(entity);
               }
            }
         }
      }
      else
      {
         Set<Serializable> ids = null;
         if (deletes == null)
         {
            deletes = new HashMap<String, Set<Serializable>>();
         }
         else
         {
            ids = deletes.get(entity);
         }
         if (ids == null)
         {
            ids = new HashSet<Serializable>();
            deletes.put(entity, ids);
         }
         ids.add(id);
      }
   }
   
   /**
    * Indicate whether there are currently any uncommitted changes (inserts,
    * updates, or deletes) in the table represented by the given DMO entity
    * name.
    * 
    * @param   entity
    *          Fully qualified name of the DMO entity to be checked.
    * 
    * @return  <code>true</code> if there any pending, uncommitted changes in
    *          the backing table, else <code>false</code>.
    */
   public boolean isEntityDirty(String entity)
   {
      int dirtyState = manager.getEntityDirtyState(entity);
      
      return (dirtyState != DirtyInfo.NONE);
   }
   
   /**
    * Indicate whether the record specified by the given entity and primary
    * key has been deleted within an uncommitted transaction within a context
    * <i>other than the current context</i>.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be checked.
    * 
    * @return  <code>true</code> if the specified record has been deleted in a
    *          different context, else <code>false</code>.
    */
   public boolean isDirtyDelete(String entity, Serializable id)
   {
      if (!manager.isDeleted(entity, id))
      {
         return false;
      }
      
      // It is marked deleted in the shared space, but if this context is the
      // one that deleted it, return false.
      Set<Serializable> deleted = null;
      if (deletes != null)
      {
         deleted = deletes.get(entity);
      }
      
      return (deleted == null || !deleted.contains(id));
   }
   
   /**
    * Retrieve from the dirty database the record specified by the given
    * entity and primary key, if it exists, and if it was not put there by the
    * current context.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key of the record to be retrieved.
    * 
    * @return  Target record if it exists in the dirty database, and was not
    *          put there by the current context, else <code>null</code>.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public Persistable getDirtyDMO(String entity, Serializable id)
   throws PersistenceException
   {
      // If we're the context making the change, we don't need the dirty
      // copy, since it's the same as the one we have in our current session.
      if (changes != null)
      {
         Set<Serializable> changed = changes.get(entity);
         if (changed != null && changed.contains(id))
         {
            return null;
         }
      }
      
      return manager.getDirtyDMO(entity, id);
   }
   
   /**
    * Execute an HQL query and return the results as a list.  If the query
    * returns multiple results per row, each element in the list will be an
    * array of <code>Object</code>s.
    * <p>
    * No locking is attempted.
    *
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   hql
    *          HQL query statement passed to Hibernate.  This method makes no
    *          assumptions as to the types of object(s) returned.
    * @param   params
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    * @param   readOnly
    *          <code>true</code> to execute the query in read-only mode, else
    *          <code>false</code>.
    *
    * @return  A list of results, or <code>null</code> if no result was found.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    */
   public List list(String entity,
                    String hql,
                    Object[] params,
                    Type[] types,
                    int maxResults,
                    int startOffset,
                    boolean readOnly)
   throws PersistenceException
   {
      int dirtyState = manager.getEntityDirtyState(entity);
      if ((dirtyState & DirtyInfo.CHANGE) == DirtyInfo.NONE)
      {
         return null;
      }
      
      return manager.list(hql,
                          params,
                          types,
                          maxResults,
                          startOffset,
                          readOnly);
   }
   
   /**
    * Given certain search criteria and possibly a potential DMO match in the
    * primary database, report any significant information currently being
    * tracked for uncommitted transactions, which might override data (or the
    * lack thereof) found in the primary database.
    * <p>
    * If a {@link DirtyInfo} object is returned by this method, it will
    * contain overriding information, such as a dirty DMO which matched the
    * search criteria, whether that record was newly inserted, or the fact
    * that the candidate record found in the primary search has been deleted
    * or modified in an uncommitted transaction.  A <code>null</code> return
    * indicates no overriding information was available.
    * 
    * @param   buffer
    *          Record buffer associated with the current query.
    * @param   index
    *          Name of index which governs the ordering of data in the current
    *          query.
    * @param   bundle
    *          An object containing one or more HQL statements to be executed
    *          against the dirty database to find a record which matches the
    *          current query criteria.  The list is in order of most specific
    *          to least specific search criteria.
    * @param   params
    *          Query substitution parameters which match placeholders in the
    *          most specific HQL statement.
    * @param   types
    *          Hibernate data types of the query substitution parameters.
    * @param   found
    *          DMO, if any, found in the primary database, which matches the
    *          current query criteria.
    * 
    * @return  An object which summarizes the findings, or <code>null</code>
    *          if no uncommitted change information was available for the
    *          given arguments.
    *          
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public DirtyInfo getDirtyInfo(RecordBuffer buffer,
                                 String index,
                                 HQLBundle bundle,
                                 Object[] params,
                                 Type[] types,
                                 Persistable found)
   throws PersistenceException
   {
      String entity = buffer.getEntityName();
      int dirtyState = manager.getEntityDirtyState(entity);
      
      // Quick out (hopefully the common case):  if there are no changes
      // pending for the target entity, we can bypass the check altogether.
      if (dirtyState == DirtyInfo.NONE)
      {
         return null;
      }
      
      Serializable foundID = (found != null ? found.getId() : null);
      
      // Has the record found in primary database been deleted?
      if ((dirtyState & DirtyInfo.DELETE) != DirtyInfo.NONE &&
          foundID != null &&
          isDirtyDelete(entity, foundID))
      {
         DirtyInfo info = new DirtyInfo();
         info.setDeleted();
         
         return info;
      }
      
      DirtyInfo info = null;
      
      if ((dirtyState & DirtyInfo.CHANGE) != DirtyInfo.NONE)
      {
         List<String> statements = bundle.getDirtyStatements();
         
         info = manager.getDirtyInfo(entity,
                                     foundID,
                                     index,
                                     statements,
                                     params,
                                     types);
         
         if (info == null)
         {
            return null;
         }
         
         Persistable dirtyDMO = info.getDirtyDMO();
         if (dirtyDMO != null && changes != null)
         {
            // Ignore changes that are driven by this context;  they are already
            // reflected in the primary database.
            Serializable id = dirtyDMO.getId();
            Set<Serializable> changed = changes.get(entity);
            if (changed != null && changed.contains(id))
            {
               return null;
            }
         }
      }
      
      return info;
   }
   
   /**
    * Clean up all information contributed to the dirty share manager by this
    * context for the current transaction.  This method is called when a full
    * database transaction has been committed or rolled back against the
    * primary database.  It is also invoked when the client context ends.  It
    * removes all records it added to the dirty database for new inserts,
    * updates, and deletes.
    * 
    * @throws  StopConditionException
    *          if a critical database error occurs that would leave the dirty
    *          database in an improper state.
    */
   public void cleanup()
   {
      try
      {
         if (changes != null)
         {
            Set<String> empty = Collections.emptySet();
            Set<String> leaked = (leakedEntities != null
                                  ? leakedEntities
                                  : empty);
            manager.cleanupChanges(changes, inserts, leaked);
         }
         if (deletes != null)
         {
            manager.cleanupDeletes(deletes);
         }
      }
      catch (PersistenceException exc)
      {
         throw new StopConditionException(exc);
      }
      finally
      {
         changes = null;
         inserts = null;
         deadInserts = null;
         deletes = null;
         dirtySnapshots = null;
         leakedEntities = null;
      }
   }
   
   /**
    * Remember that the record represented by the given entity and primary key
    * ID has been modified in the current context.  This is necessary for
    * proper cleanup at the end of the current transaction.
    * 
    * @param   entity
    *          Name of an entity, which is the name of the DMO implementation
    *          class associated with a table being tracked for uncommitted
    *          changes.
    * @param   id
    *          Primary key ID of the modified record.
    */
   private void trackChange(String entity, Serializable id)
   {
      if (changes == null)
      {
         changes = new HashMap<String, Set<Serializable>>();
      }
      
      if (dirtySnapshots == null)
      {
         dirtySnapshots = new HashSet<RecordIdentifier>();
      }
      
      String _entity = entity.intern();
      Set<Serializable> ids = changes.get(_entity);
      if (ids == null)
      {
         ids = new HashSet<Serializable>();
         changes.put(_entity, ids);
      }
      ids.add(id);
      
      RecordIdentifier ident = new RecordIdentifier(_entity, id);
      dirtySnapshots.add(ident);
   }
}
