/*
** Module   : RecordLockContext.java
** Abstract : Coordinates record locks across record buffers within a session
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20070411   @32990 Created initial version. Coordinates record
**                           locks across record buffers within a session.
** 002 ECF 20070414   @32995 Fixed ConcurrentModificationException. Cannot
**                           modify lockStates map while iterating it
**                           except via iterator. Changed collection type
**                           for buffers in LockState to LinkedList from
**                           HashSet to conserve memory.
** 003 SVL 20080512   @38259 Added support for the
**                           RecordBuffer.pinnedLockTypes map.
** 004 ECF 20080821   @39564 Disallow exclusive lock to be acquired unless
**                           a transaction is active. Normally, this will
**                           not occur in converted code. It is added as a
**                           failsafe for hand-written code.
** 005 ECF 20081114   @40454 Fixed record lock leak. Added implementations
**                           of hashCode() and equals() to LockState, to
**                           ensure duplicates are not stored in the hash
**                           sets of LockStates in the locksByBuffer map.
**                           Integrated generics.
** 006 ECF 20090202   @41250 Fixed LockState storage. We now store record
**                           identifiers by buffer instead of LockState
**                           objects by buffer. This prevents duplicate,
**                           conflicting LockState objects from corrupting
**                           state.
** 007 ECF 20090218   @41333 Fixed memory leak. Perm.locksByBuffer is now
**                           a WeakHashMap to prevent RecordBuffer objects
**                           from being pinned in memory.
** 008 SVL 20090708   @43113 Into endTransaction(), if the given record
**                           isn't loaded into any corresponding buffer,
**                           downgrade record's lock to NONE.
** 009 ECF 20090824   @43732 Fixed Perm.relinquishBufferLocks(). We were
**                           leaking RecordIdentifiers in the locksByBuffer
**                           collection, which was causing both a memory leak
**                           and a performance problem.
** 010 ECF 20131028          Moved to lock sub-package.
*/

package com.goldencode.p2j.persist.lock;

import java.io.*;
import java.util.*;
import org.apache.commons.logging.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;

/**
 * An instance of this class coordinates record locking activity across all
 * record buffers for a particular database, within a user session.  While
 * the {@link LockManager} handles the actual synchronization and acquisition
 * of record locks across all sessions for a given database, an instance of
 * <code>RecordLockContext</code> is responsible for ensuring that all record
 * buffers which access a given record lock <i>within a session</i> cooperate
 * properly in their interactions with the lock manager.
 * <p>
 * When a record lock is required, it generally is associated with one or
 * more record buffers in a session.  There are several housekeeping
 * processes in the persistence framework which may temporarily require a
 * lock, but generally, record locks are associated with record buffers as
 * the result of some converted Progress language statement option which
 * (possibly implicitly) determines a record lock (i.e., NO-LOCK, SHARE-LOCK,
 * EXCLUSIVE-LOCK on FIND, GET, FOR, OPEN QUERY, CREATE, etc.).
 * <p>
 * In P2J, this association is made by {@link RecordBuffer} at the time it
 * creates a new record, or when a record found by a database query is stored
 * in the buffer via the {@link RecordBuffer#setRecord setRecord method}.
 * <p>
 * A record lock is associated with a single record;  there can be only one
 * type of lock granted for a particular combination of database, table, and
 * primary key identifier at a time across all sessions.  However, multiple
 * record buffers which contain the same record at the same time <i>within a
 * session</i> all share access to the lock, such that each of them are
 * granted the set of rights (or lack thereof) associated with the lock.
 * Since each session is essentially single threaded, this is safe.
 * <p>
 * Record locking across buffers within a session is asymmetric:  any buffer
 * can acquire and upgrade the communal lock, but all buffers must agree
 * before the communal lock can be downgraded or released.  More specifically,
 * each buffer maintains its own state about its particular requirement for
 * the state of the lock (its "pinned" lock type -- the lock type which will
 * potentially "pin" the state of the communal lock at a certain, minimum
 * level until released by that buffer).  If a buffer's pinned lock type is
 * <code>LockType.NONE</code>, then that buffer effectively does not pin the
 * communal lock at all, and the lock may be released if no other buffers are
 * pinning it to a higher level.  However, if a buffer's pinned lock type is
 * <code>LockType.EXCLUSIVE</code> or <code>LockType.SHARE</code>, the
 * communal lock may not be downgraded or released, respectively, until that
 * buffer changes its pinned type to a lower level.
 * <p>
 * In addition to this requirement for buffers to cooperate for lock
 * downgrade and release, a SHARE lock may not be released, nor an EXCLUSIVE
 * lock downgraded, while an application level transaction is active.  The
 * presence of a transaction trumps the pinning rules noted above.  Once the
 * transaction ends, via commit or rollback, those rules apply again.
 * <p>
 * This class is abstract, and two different subclasses, both defined here,
 * are used to support permanent database tables and temporary tables.  The
 * factory method {@link #get} is used to acquire an instance of the
 * appropriate type.  It should only be invoked once per database, per
 * session and is only intended for use by the {@link Persistence} class.
 */
public abstract class RecordLockContext
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(RecordLockContext.class);
   
   /**
    * Factory method used to retrieve an instance of a
    * <code>RecordLockContext</code> implementation which is appropriate for
    * the database whose records are being managed.  Should only be called by
    * the {@link Persistence} class.
    * <p>
    * The implementation returned for use with temporary tables essentially
    * is a do-nothing implementation, since record locking is a no-op for
    * temp tables.
    * <p>
    * The implementation returned for permanent tables adheres to the rules
    * noted in the class description above.
    * <p>
    * This dual architecture is used so that other classes which use the
    * services of this class do not have to differentiate their behavior for
    * temporary vs. permananet tables.  This permits client classes to remain
    * unaware of the implementation differences for permanent and temporary
    * tables.
    *
    * @param   persistence
    *          Persistence service object which will be used for access to
    *          the global lock manager.
    * @param   bufferManager
    *          Buffer manager which tracks buffer scopes.
    *
    * @return  An instance of this class.
    */
   public static RecordLockContext get(Persistence persistence, BufferManager bufferManager)
   {
      if (DatabaseManager.TEMP_TABLE_DB.equals(persistence.getDatabase()))
      {
         return new RecordLockContext()
         {
         };
      }
      
      return new Perm(persistence, bufferManager);
   }
   
   /**
    * Default constructor accessible only from within this class.
    */
   private RecordLockContext()
   {
   }
   
   /**
    * Register a record buffer to a particular record lock, which must
    * have been created previously.  This creates a bidirectional, one to
    * many association between the buffer and the record lock.
    * <p>
    * Explicit deregistration of this association is unnecessary.  It
    * occurs implicitly when the lock is released and/or the buffer goes
    * out of scope.
    * <p>
    * This implementation does nothing.  It should be overridden for a
    * meaningful implementation.
    *
    * @param   ident
    *          Unique identifier for a particular database record.
    * @param   buffer
    *          Record buffer to be associated with the specified record
    *          lock.
    */
   public void registerBuffer(RecordIdentifier ident, RecordBuffer buffer)
   {
   }
   
   /**
    * Attempt to release all locks associated with the given buffer.
    * Release of any given lock may be deferred if other buffers in the
    * session are pinning the lock.
    * <p>
    * This implementation does nothing.  It should be overridden for a
    * meaningful implementation.
    *
    * @param   buffer
    *          Record buffer whose locks are being relinquished.
    */
   public void relinquishBufferLocks(RecordBuffer buffer)
   {
   }
   
   /**
    * Process any pending lock changes which have been deferred during the
    * most recent application level transaction.  Specifically, release or
    * downgrade any locks which are no longer being pinned by record
    * buffers.
    * <p>
    * This implementation does nothing.  It should be overridden for a
    * meaningful implementation.
    */
   public void transactionEnded()
   {
   }
   
   /**
    * Request that the communal lock on a record be acquired or otherwise
    * modified to match the requested type.
    * <p>
    * This implementation does nothing.  It should be overridden for a
    * meaningful implementation.
    *
    * @param   ident
    *          Unique identifier for a particular database record.
    * @param   requestedType
    *          The lock type being requested.
    * @param   buffer
    *          Record buffer on whose behalf the request is being made.
    *          May be <code>null</code> in cases where the association
    *          will be made later or if the lock is being managed for
    *          framework housekeeping purposes.
    *
    * @return  The actual lock type held as a result of the invocation.
    *
    * @throws  LockUnavailableException
    *          never.
    */
   public LockType lock(RecordIdentifier ident, LockType requestedType, RecordBuffer buffer)
   throws LockUnavailableException
   {
      return LockType.NONE;
   }
   
   /**
    * Request that the communal lock on a record be acquired or otherwise
    * modified to match the requested type.  If a record buffer is
    * provided, its pinned lock type is adjusted to match the request.
    * <p>
    * The communal lock may or may not be modified immediately, according
    * to the rules described in this class' description.  Generally, a
    * request to upgrade or acquire the lock for the first time is honored
    * immediately.  However, a request to downgrade or release the lock
    * may be deferred.
    *
    * @param   ident
    *          Unique identifier for a particular database record.
    * @param   requestedType
    *          The lock type being requested.
    *
    * @return  The actual lock type held as a result of the invocation.
    *
    * @throws  LockUnavailableException
    *          if a no-wait lock variant is requested, and another session
    *          currently holds the lock.  This will never be thrown in
    *          cases where a downgrade or release is being requested.
    */
   public LockType lock(RecordIdentifier ident, LockType requestedType)
   throws LockUnavailableException
   {
      return lock(ident, requestedType, null);
   }
   
   /**
    * Concrete implementation of <code>RecordLockContext</code> which enables
    * intra-session record lock sharing across record buffers for permanent
    * database tables.
    */
   private static class Perm
   extends RecordLockContext
   {
      /** Persistence service object through which lock manager is accessed */
      private final Persistence persistence;
      
      /** Buffer manager which tracks buffer scopes */
      private final BufferManager bufferManager;
      
      /** Mapping of record identifiers to lock states */
      private final Map<RecordIdentifier, LockState> lockStates =
         new HashMap<RecordIdentifier, LockState>();
      
      /** Mapping of buffers to lock state sets */
      private final Map<RecordBuffer, Set<RecordIdentifier>> locksByBuffer =
         new WeakHashMap<RecordBuffer, Set<RecordIdentifier>>();
      
      /**
       * Constructor accessible only from within this class.
       *
       * @param   persistence
       *          Persistence service object which will be used for access to
       *          the global lock manager.
       * @param   bufferManager
       *          Buffer manager which tracks buffer scopes.
       */
      private Perm(Persistence persistence, BufferManager bufferManager)
      {
         super();
         
         this.persistence = persistence;
         this.bufferManager = bufferManager;
      }
      
      /**
       * Register a record buffer to a particular record lock, which must
       * have been created previously.  This creates a bidirectional, one to
       * many association between the buffer and the record lock.
       * <p>
       * Explicit deregistration of this association is unnecessary.  It
       * occurs implicitly when the lock is released and/or the buffer goes
       * out of scope.
       *
       * @param   ident
       *          Unique identifier for a particular database record.
       * @param   buffer
       *          Record buffer to be associated with the specified record
       *          lock.
       */
      public void registerBuffer(RecordIdentifier ident, RecordBuffer buffer)
      {
         LockState state = getLockState(ident, false);
         
         if (state == null)
         {
            return;
         }
         
         if (state.registerBuffer(buffer))
         {
            Set<RecordIdentifier> locks = locksByBuffer.get(buffer);
            if (locks == null)
            {
               locks = new HashSet<RecordIdentifier>(4);
               locksByBuffer.put(buffer, locks);
            }
            locks.add(ident);
         }
      }
      
      /**
       * Request that the communal lock on a record be acquired or otherwise
       * modified to match the requested type.  If a record buffer is
       * provided, the pinned lock type of the specified record is adjusted to
       * match the request.
       * <p>
       * The communal lock may or may not be modified immediately, according
       * to the rules described in this class' description.  Generally, a
       * request to upgrade or acquire the lock for the first time is honored
       * immediately.  However, a request to downgrade or release the lock
       * may be deferred.
       *
       * @param   ident
       *          Unique identifier for a particular database record.
       * @param   requestedType
       *          The lock type being requested.
       * @param   buffer
       *          Record buffer on whose behalf the request is being made.
       *          May be <code>null</code> in cases where the association
       *          will be made later or if the lock is being managed for
       *          framework housekeeping purposes.
       *
       * @return  The actual lock type held as a result of the invocation.
       *
       * @throws  LockUnavailableException
       *          if a no-wait lock variant is requested, and another session
       *          currently holds the lock.  This will never be thrown in
       *          cases where a downgrade or release is being requested.
       * @throws  IllegalStateException
       *          if an exclusive lock is requested outside an active
       *          transaction.
       */
      public LockType lock(RecordIdentifier ident, LockType requestedType, RecordBuffer buffer)
      throws LockUnavailableException
      {
         if (buffer != null)
         {
            // Set lock type buffer currently is pinning to requested type.
            buffer.setPinnedLockType(ident.getRecordID(), requestedType);
         }
         
         boolean create = (requestedType != LockType.NONE);
         LockState state = getLockState(ident, create);
         
         if (state == null)
         {
            return LockType.NONE;
         }
         
         LockType currentType = state.getCurrentType();
         int comparison = currentType.compareTo(requestedType);
         
         if (comparison < 0)
         {
            // Current type is less restrictive than requested type;  attempt
            // to upgrade lock immediately.
            
            // Disallow exclusive lock outside a transaction.
            if (requestedType.isExclusive() && !TransactionManager.isTransaction())
            {
               throw new IllegalStateException(
                  "Cannot acquire an exclusive lock outside a transaction (" +
                  ident +
                  ")");
            }
            
            // Perform the upgrade.
            persistence.lock(requestedType, ident, true);
            state.setCurrentType(requestedType);
         }
         else if (comparison > 0)
         {
            // Current type is more restrictive than requested type;  must
            // follow rules to downgrade or release lock.
            state.relinquish(requestedType);
         }
         
         currentType = state.getCurrentType();
         if (currentType == LockType.NONE)
         {
            lockStates.remove(ident);
         }
         
         return currentType;
      }
      
      /**
       * Attempt to release all locks associated with the given buffer.
       * Release of any given lock may be deferred if other buffers in the
       * session are pinning the lock.
       *
       * @param   buffer
       *          Record buffer whose locks are being relinquished.
       */
      public void relinquishBufferLocks(RecordBuffer buffer)
      {
         Set<RecordIdentifier> locks = locksByBuffer.get(buffer);
         
         if (locks == null)
         {
            return;
         }
         
         try
         {
            Iterator<RecordIdentifier> iter = locks.iterator();
            while (iter.hasNext())
            {
               RecordIdentifier ident = iter.next();
               LockState state = getLockState(ident, false);
               if (state != null)
               {
                  state.relinquish(LockType.NONE);
                  if (state.getCurrentType() == LockType.NONE)
                  {
                     lockStates.remove(ident);
                     iter.remove();
                  }
               }
               else
               {
                  iter.remove();
               }
            }
         }
         catch (LockUnavailableException exc)
         {
            // Never thrown.
         }
         finally
         {
            if (locks.isEmpty())
            {
               locksByBuffer.remove(buffer);
            }
         }
      }
      
      /**
       * Process any pending lock changes which have been deferred during the
       * most recent application level transaction.  Specifically, release or
       * downgrade any locks which are no longer being pinned by record buffers.
       */
      public void transactionEnded()
      {
         try
         {
            Iterator<LockState> iter = lockStates.values().iterator();
            while (iter.hasNext())
            {
               LockState state = iter.next();
               state.transactionEnded();
               if (state.getCurrentType() == LockType.NONE)
               {
                  iter.remove();
               }
            }
         }
         catch (LockUnavailableException exc)
         {
            // Never thrown.
         }
      }
      
      /**
       * Get the <code>LockState</code> object associated with the given
       * record identifier, optionally creating it if it does not yet exist.
       *
       * @param   ident
       *          Unique identifier for a particular database record.
       * @param   create
       *          If <code>true</code>, create the lock state object if not
       *          found;  if <code>false</code>, do not create it.
       * 
       * @return  <code>LockState</code> for given identifier.
       */
      private LockState getLockState(RecordIdentifier ident, boolean create)
      {
         LockState state = lockStates.get(ident);
         if (state == null && create)
         {
            state = new LockState(ident);
            lockStates.put(ident, state);
         }
         
         return state;
      }
      
      /**
       * The lock state of a particular record, which includes its current
       * lock type, its unique identifier, and the set of buffers which share
       * access to the backing lock.
       */
      private final class LockState
      {
         /** Database record whose lock state is being tracked */
         private final RecordIdentifier ident;
         
         /** Buffers who share this lock state */
         private final List<RecordBuffer> buffers =
            new LinkedList<RecordBuffer>();
         
         /** Lock type currently assigned to the record */
         private LockType currentType = null;
         
         /**
          * Constructor which queries current lock type for a record from the
          * lock manager.
          *
          * @param   ident
          *          Unique identifier for the database record whose lock
          *          type is to be queried from the lock manager.
          */
         LockState(RecordIdentifier ident)
         {
            this.ident = ident;
            this.currentType = persistence.queryLock(ident);
         }
         
         /**
          * Override parent's implementation to be consistent with {@link
          * #equals(Object)}.
          * 
          * @return  Hash code for this object.
          */
         public int hashCode()
         {
            return (37 * 17 + ident.hashCode());
         }
         
         /**
          * Override parent's implementation of equality.  Two instances of
          * this class are considered equal if they represent the same
          * lockable resource, regardless of the other internal state.
          * 
          * @param   o
          *          Another instance of this class to test for equality with
          *          this instance.
          * 
          * @return  <code>true</code> if this object and the given instance
          *          have the same record identifier, else <code>false</code>.
          */
         public boolean equals(Object o)
         {
            return ident.equals(((LockState) o).ident);
         }
         
         /**
          * Register a record buffer which shares access to the record's
          * current lock.  Deregistration occurs implicitly when this buffer
          * no longer pins the lock.
          *
          * @param   buffer
          *          Record buffer to be registered.
          *
          * @return  <code>true</code> if the buffer was added;
          *          <code>false</code> if it was previously registered.
          */
         boolean registerBuffer(RecordBuffer buffer)
         {
            if (!buffers.contains(buffer))
            {
               buffers.add(buffer);
               
               return true;
            }
            
            return false;
         }
         
         /**
          * Get record's unique identifier.
          *
          * @return  Unique identifier.
          */
         RecordIdentifier getIdentifier()
         {
            return ident;
         }
         
         /**
          * Get the current lock type.
          *
          * @return  Current lock type.
          */
         LockType getCurrentType()
         {
            return currentType;
         }
         
         /**
          * Set the current lock type.  This will always set the current type
          * to a <i>non-</i>no-wait variant.
          *
          * @param   currentType
          *          Current type to be set.  Only the category is used
          *          (i.e., NONE, SHARE, EXCLUSIVE);  the no-wait-"ness" of
          *          the lock type is discarded.
          */
         void setCurrentType(LockType currentType)
         {
            this.currentType = currentType.toWaitVariant();
         }
         
         /**
          * Process any pending lock change which has been deferred during
          * the most recent application level transaction.  Specifically,
          * release or downgrade a lock which is no longer being pinned by
          * record buffers.
          *
          * @throws  LockUnavailableException
          *          never.
          */
         void transactionEnded()
         throws LockUnavailableException
         {
            if (currentType == LockType.NONE)
            {
               return;
            }
            
            int blockDepth = bufferManager.getCurrentScope();
            LockType maxLock = LockType.NONE;
            
            Iterator<RecordBuffer> iter = buffers.iterator();

            boolean loaded = false;
            while (iter.hasNext())
            {
               RecordBuffer buffer = iter.next();
               Persistable record = buffer.getCurrentRecord();
               if (record != null &&
                   record.getId().equals(ident.getRecordID()))
               {
                  loaded = true;
                  break;
               }
            }
            
            if (loaded)
            {
               iter = buffers.iterator();
               while (iter.hasNext())
               {
                  RecordBuffer buffer = iter.next();
                  
                  Serializable id = ident.getRecordID();
                  LockType pinned = buffer.getPinnedLockType(id);
                  if (pinned == LockType.NONE)
                  {
                     continue;
                  }
                  
                  // Scope depth at which this buffer's scope was first opened.
                  int activeScopeDepth = buffer.getScopeOpenDepth();
                  
                  if (activeScopeDepth > 0 && blockDepth > activeScopeDepth)
                  {
                     // Buffer scope pinning this record lock is still open AND
                     // we're not yet in the block scope in which the buffer
                     // scope will end.  Downgrade to (or maintain) share lock.
                     pinned = LockType.SHARE;
                  }
                  else
                  {
                     // Buffer scope is closed.  Stop pinning the lock.
                     pinned = LockType.NONE;
                  }
                  
                  buffer.setPinnedLockType(id, pinned);
                  
                  // Remember the most restrictive lock type pinned by any
                  // buffer.
                  if (pinned.compareTo(maxLock) > 0)
                  {
                     maxLock = pinned;
                  }
               }
               
               // Downgrade or remove lock context-wide.
               updateLock(maxLock);
            }
            else
            {
               updateLock(LockType.NONE);
            }
         }
         
         /**
          * Attempt to relinquish the lock represented by this object,
          * releasing or downgrading it if possible.  A change is only
          * possible outside a transaction, and then only to the most
          * restrictive lock pinned by any buffer sharing the communal lock.
          *
          * @param   requestedType
          *          NONE or SHARE only (it doesn't make sense to downgrade
          *          to EXCLUSIVE).
          *
          * @throws  IllegalArgumentException
          *          if an EXCLUSIVE lock is requested.
          * @throws  LockUnavailableException
          *          never, since we never request a no-wait lock here.
          */
         void relinquish(LockType requestedType)
         throws LockUnavailableException
         {
            if (requestedType.isExclusive())
            {
               throw new IllegalArgumentException(
                  "Cannot downgrade lock type to EXCLUSIVE");
            }
            
            // If there is no change to be made, or if we are in an active
            // transaction, nothing more to do.
            if (currentType == requestedType.toWaitVariant() ||
                TransactionManager.isTransaction())
            {
               return;
            }
            
            // Walk through buffers and determine whether any one would pin
            // the lock to SHARE.  If not, release the lock, otherwise make
            // it SHARE.
            // NOTE:  Since we are outside of a transaction, no buffer will
            // be pinning to EXCLUSIVE.  Also, requestedType will only ever
            // be NONE or SHARE, so this algorithm is safe.
            Iterator<RecordBuffer> iter = buffers.iterator();
            while (iter.hasNext())
            {
               RecordBuffer buffer = iter.next();
               LockType pinned =
                  buffer.getPinnedLockType(ident.getRecordID());
               
               if (pinned == LockType.NONE)
               {
                  // Buffer no longer pins the lock, so remove its
                  // association.
                  iter.remove();
               }
               else
               {
                  // Can't go below SHARE.
                  requestedType = LockType.SHARE;
               }
            }
            
            updateLock(requestedType);
         }
         
         /**
          * Attempt to update the current lock on the record immediately.
          *
          * @param   lockType
          *          Requested lock type.  If NONE, this object will be
          *          removed from the map of lock states by identifier.
          *
          * @throws  LockUnavailableException
          *          never.
          */
         private void updateLock(LockType lockType)
         throws LockUnavailableException
         {
            if (lockType.toWaitVariant() != currentType)
            {
               persistence.lock(lockType, ident, true);
               setCurrentType(lockType);
            }
         }
      }
   }
}
