/*
** Module   : RecordLockEvent.java
** Abstract : Object which stores information about a record lock event
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20131101 Created initial version.
*/

package com.goldencode.p2j.persist.lock;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

/**
 * Object which stores information about a record lock event.
 */
public final class RecordLockEvent
{
   /** Sequential, unique event ID */
   private final long eventID;
   
   /** The locker */
   private final SessionToken locker;
   
   /** Unique identifier of the locked record */
   private final RecordIdentifier identifier;
   
   /** Lock type before the event */
   private final LockType oldType;
   
   /** Lock type after the event */
   private final LockType newType;
   
   /** Unique, numeric identifier of user in whose session the event occurred */
   private integer userid = null;
   
   /**
    * Constructor.
    * 
    * @param   eventID
    *          Sequential, unique event ID.
    * @param   locker
    *          The locker.
    * @param   identifier
    *          Unique identifier of the locked record.
    * @param   oldType
    *          Lock type before the event.
    * @param   newType
    *          Lock type after the event.
    */
   RecordLockEvent(long eventID,
                   SessionToken locker,
                   RecordIdentifier identifier,
                   LockType oldType,
                   LockType newType)
   {
      this.eventID = eventID;
      this.locker = locker;
      this.identifier = identifier;
      this.oldType = oldType;
      this.newType = newType;
   }
   
   /**
    * Get the unique ID of this event.
    * 
    * @return  Event ID.
    */
   public long getEventID()
   {
      return eventID;
   }
   
   /**
    * Get the session token of the user which generated this lock event.
    * 
    * @return  Locker's session token.
    */
   public SessionToken getLocker()
   {
      return locker;
   }
   
   /**
    * Get the identifier which uniquely identifies the locked record.
    * 
    * @return  Locked record identifier.
    */
   public RecordIdentifier getIdentifier()
   {
      return identifier;
   }
   
   /**
    * Get the lock type as it was before this event.
    * 
    * @return  Previous lock type.
    */
   public LockType getOldType()
   {
      return oldType;
   }
   
   /**
    * Get the lock type as it was after this event.
    * 
    * @return  New lock type.
    */
   public LockType getNewType()
   {
      return newType;
   }
   
   /**
    * Get the ID of the user who generated this lock event.
    * 
    * @return  User ID.
    */
   public integer getUserid()
   {
      return userid;
   }
   
   /**
    * Set the ID of the user who generated this lock event.
    * @param userid
    */
   public void setUserid(integer userid)
   {
      this.userid = userid;
   }
   
   /**
    * Get a string describing this event.
    * 
    * @return  String representation of this event.
    */
   @Override
   public String toString()
   {
      StringBuilder buf = new StringBuilder("Lock event: ");
      buf.append(locker.toString());
      buf.append(" ");
      buf.append(identifier.toString());
      buf.append("; ");
      buf.append(oldType.toString());
      buf.append("->");
      buf.append(newType.toString());
      
      return buf.toString();
   }
}
