/*
** Module   : LockManagerMultiplexerImpl.java
** Abstract : Handles server-to-server requests for record lock management
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- --------------------------Description---------------------------
** 001 ECF 20071018 ADD  @35553  Created initial version. Handles server to
**                               server requests for record lock management.
** 002 ECF 20071016 CHG  @35911  Completed implementation of the initialize()
**                               method. Registers the singleton instance of
**                               this class as a network server object.
** 003 ECF 20071128 CHG  @36055  Replace LockType with int for method return
**                               values and parameters. This prevents us from
**                               having to support LockType as Serializable
**                               objects, which is desirable for several
**                               reasons:
**                               1) less data to send over the network;
**                               2) allows us to avoid creating multiple
**                                  instances of each LockType type, avoiding
**                                  a lot of unnecessary working set.
** 004 VMN 20130830              Added lockTable method.
** 005 ECF 20131028          Import change.
*/

package com.goldencode.p2j.persist.remote;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.lock.*;

/**
 * A singleton instance of this class services server-to-server requests for
 * database record lock management.  This instance is registered as a remote
 * object and serves record lock management requests for one or more databases
 * which the server authoritatively manages.
 * <p>
 * Service requests may originate from more than one external server, so this
 * implementation is threadsafe.
 * 
 * @author ECF
 */
public final class LockManagerMultiplexerImpl
extends DatabaseMultiplexer<LockManager>
implements LockManagerMultiplexer
{
   /** Singleton instance of this class */
   private static LockManagerMultiplexer instance = null;
   
   /**
    * Initialize the lock manager multiplexer for this server instance.  This
    * method is to be called during server bootstrap, but not before the local
    * database manager has been initialized.
    * <p>
    * This method enforces the singleton pattern.
    * 
    * @throws  IllegalStateException
    *          if this method is invoked more than once.
    */
   public static synchronized void initialize()
   {
      if (instance != null)
      {
         throw new IllegalStateException(
            "Lock manager multiplexer already initialized for this process");
      }
      
      instance = new LockManagerMultiplexerImpl();
      RemoteObject.registerNetworkServer(LockManagerMultiplexer.class,
                                         instance);
   }
   
   /**
    * Default constructor which initializes the resources needed for this
    * multiplexer to service lock manager requests.
    */
   private LockManagerMultiplexerImpl()
   {
      super();
   }
   
   /**
    * Attempt to obtain the specified lock type on a particular record.  If
    * the lock cannot be obtained immediately, the current thread will block
    * until the lock becomes available and is obtained, unless the requested
    * lock type is a "no-wait" lock.  In the latter case, an exception is
    * raised, indicating that the lock is not available.  The current thread
    * continues once the lock is obtained.
    * <p>
    * A lock currently held can be released by specifying
    * <code>LockType.NONE</code> for the <code>lockType</code> parameter.
    * <p>
    * The normal return of this method indicates that the lock request (or
    * lack thereof) completed successfully, and that the current context now
    * holds a lock of the requested type.
    *
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   lockType
    *          Type of lock to be obtained.  <code>LockType.NONE</code> is
    *          used to release an existing lock, and to continue with no
    *          lock.
    * @param   ident
    *          ID which uniquely identifies the record being locked.
    * @param   update
    *          <code>true</code> if the lock state for the current record
    *          should be updated;  else <code>false</code>.
    *
    * @throws  LockUnavailableException
    *          if a "no-wait" lock cannot be acquired immediately.
    */
   public void lock(int database,
                    int lockType,
                    RecordIdentifier ident,
                    boolean update)
   throws LockUnavailableException
   {
      LockType lt = LockType.fromInt(lockType);
      lookupWorker(database).lock(lt, ident, update);
   }

   /**
    * Attempt to obtain the specified lock type on a table. If the lock cannot be obtained
    * immediately, the current thread will block until the lock becomes available and is
    * obtained, unless the requested lock type is a "no-wait" lock.  In the latter case,
    * an exception is raised, indicating that the lock is not available. The current thread
    * continues once the lock is obtained.
    * <p>
    * A lock currently held can be released by specifying
    * <code>LockType.NONE</code> for the <code>lockType</code> parameter.
    * <p>
    * Method returns previous lock type (so that calling code knows what to set it back to after
    * bulk delete).
    * The normal return of this method indicates that the lock request (or lack thereof)
    * completed successfully, and that the current context now holds a lock of the requested
    * type.
    *
    * @param   database
    *          Value that uniquely identifies the target database on the remote server.
    * @param   lockType
    *          Type of lock to be obtained.  <code>LockType.NONE</code> is used to release
    *          an existing lock, and to continue with no lock.
    * @param   table
    *          Table name.
    * @param   update
    *          <code>true</code> if the lock state for the current record should be updated;
    *          else <code>false</code>.
    *
    * @throws  LockUnavailableException
    *          if a "no-wait" lock cannot be acquired immediately.
    *
    * @return  previous lock type.
    */
   public LockType lockTable(int database,
                             int lockType,
                             String table,
                             boolean update)
   throws LockUnavailableException
   {
      LockType lt = LockType.fromInt(lockType);
      return lookupWorker(database).lockTable(lt, table, update);
   }

   /**
    * Query the lock type currently held by the current context for the
    * specified database record.
    * <p>
    * This method should always return the <b>non-</b><code>NO_WAIT</code>
    * variants of a lock type.  That is, even if the actual lock type is
    * <code>LockType.EXCLUSIVE_NO_WAIT</code> or
    * <code>LockType.SHARE_NO_WAIT</code>, the simpler types of
    * <code>LockType.EXCLUSIVE</code> and <code>LockType.SHARE</code>,
    * respectively, should be returned.  The lack of a lock is required to
    * return <code>LockType.NONE</code> rather than <code>null</code>.
    *
    * @param   database
    *          Value that uniquely identifies the target database on the
    *          remote server.
    * @param   ident
    *          ID which uniquely identifies the record being queried.
    *
    * @return  The lock type currently held by the current context, never
    *          <code>null</code>.
    */
   public int queryLock(int database, RecordIdentifier ident)
   {
      LockType lt = lookupWorker(database).queryLock(ident);
      
      return LockType.toInt(lt);
   }
   
   /**
    * Get the lock manager which is associated with the given database.
    * 
    * @param   database
    *          Physical database with which the returned worker is associated.
    *          
    * @return  Worker object to which service requests associated with the
    *          given database must be delegated.
    */
   protected LockManager getWorker(Database database)
   {
      Persistence p = PersistenceFactory.getInstance(database);
      
      return p.getLockManager();
   }
}
