/*
** Module   : RemoteLockManager.java
** Abstract : LockManager implementation which delegates to a remote server
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20071018  @35555  Created initial version. LockManager
**                           implementation which delegates to a remote
**                           server.
** 002 ECF 20071128  @36056  Modified calls to LockManagerMultiplexer to
**                           comply with API changes.
** 003 ECF 20090801  @43501  Added getLockAdminstrator(). Always returns null,
**                           since administration of remote lock managers is
**                           not supported.
** 004 VMN 20130830          Added lockTable method.
** 005 ECF 20131031          Added setLockListener method to satisfy LockManager interface, but it
**                           is not supported in this remote implementation.
*/

package com.goldencode.p2j.persist.remote;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.lock.*;

/**
 * A specialized implementation of {@link
 * com.goldencode.p2j.persist.lock.LockManager LockManager} which delegates
 * requests for database record locking services to the remote P2J server
 * which is authoritative for the target database.  This is necessary in cases
 * where business logic dynamically connects to a remote database which is not
 * managed by the current P2J server instance, since record lock actions must
 * be serialized through a lock manager which is authoritative for the target
 * database.
 * <p>
 * An instance of this class works with a local proxy for a remote instance of
 * {@link LockManagerMultiplexer} residing at the target server.  The remote
 * object may manage multiple databases, each of which is assigned a unique
 * database identifier.  When the database is {@link #setDatabase(Database)
 * set} into a <code>RemoteLockManager</code> instance, the associated
 * identifier is requested from the remote multiplexer.  The ID returned by
 * the multiplexer is used on all subsequent service requests sent by the
 * <code>RemoteLockManager</code> instance.
 * 
 * @see     RemotePersistence
 * @author  ECF
 */
public final class RemoteLockManager
implements LockManager
{
   /** Lock manager multiplexer which will service remote requests */
   private final LockManagerMultiplexer multiplexer;
   
   /** Multiplex ID for the remote database associated with this object */
   private int mpID = -1;
   
   /**
    * Constructor which stores a proxy to the remote object which will
    * service lock manager requests for the remote database.
    * 
    * @param   multiplexer
    *          Lock manager multiplexer which will service remote requests.
    */
   public RemoteLockManager(LockManagerMultiplexer multiplexer)
   {
      this.multiplexer = multiplexer;
   }
   
   /**
    * Set the name of the physical database with which this lock manager is
    * associated.
    *
    * @param   database
    *          Database information object.
    */
   public void setDatabase(Database database)
   {
      if (mpID >= 0)
      {
         throw new IllegalStateException("Database can only be set once");
      }
      
      mpID = multiplexer.getMultiplexID(database.getName());
   }
   
   /**
    * Register a lock listener, which will receive notifications of changes in lock status for any
    * record managed by this object.
    * <p>
    * Remote lock update notifications are not supported, so this method will throw an exception
    * if invoked.
    * 
    * @param   listener
    *          Listener to be registered or <code>null</code> to deregister the current listener.
    * 
    * @throws  UnsupportedOperationException
    *          always.
    */
   public void setLockListener(LockListener listener)
   {
      throw new UnsupportedOperationException();
   }
   
   /**
    * Retrieve the {@link LockAdministrator} object associated with this
    * <code>LockManager</code> instance.
    * <p>
    * This implementation returns <code>null</code>;  the administration of a
    * remote lock manager is not supported.
    * 
    * @return  <code>null</code>.
    */
   public LockAdministrator getLockAdministrator()
   {
      return null;
   }
   
   /**
    * Attempt to obtain the specified lock type on a particular record.  If
    * the lock cannot be obtained immediately, the current thread will block
    * until the lock becomes available and is obtained, unless the requested
    * lock type is a "no-wait" lock.  In the latter case, an exception is
    * raised, indicating that the lock is not available.  The current thread
    * continues once the lock is obtained.
    * <p>
    * A lock currently held can be released by specifying
    * <code>LockType.NONE</code> for the <code>lockType</code> parameter.
    * <p>
    * The normal return of this method indicates that the lock request (or
    * lack thereof) completed successfully, and that the current context now
    * holds a lock of the requested type.
    *
    * @param   lockType
    *          Type of lock to be obtained.  <code>LockType.NONE</code> is
    *          used to release an existing lock, and to continue with no
    *          lock.
    * @param   ident
    *          ID which uniquely identifies the record being locked.
    * @param   update
    *          <code>true</code> if the lock state for the current record
    *          should be updated;  else <code>false</code>.
    *
    * @throws  LockUnavailableException
    *          if a "no-wait" lock cannot be acquired immediately.
    */
   public void lock(LockType lockType, RecordIdentifier ident, boolean update)
   throws LockUnavailableException
   {
      int type = LockType.toInt(lockType);
      multiplexer.lock(mpID, type, ident, update);
   }

   /**
    * Attempt to obtain the specified lock type on a table. If the lock cannot be obtained
    * immediately, the current thread will block until the lock becomes available and is
    * obtained, unless the requested lock type is a "no-wait" lock.  In the latter case,
    * an exception is raised, indicating that the lock is not available. The current thread
    * continues once the lock is obtained.
    * <p>
    * A lock currently held can be released by specifying
    * <code>LockType.NONE</code> for the <code>lockType</code> parameter.
    * <p>
    * Method returns previous lock type (so that calling code knows what to set it back to after
    * bulk delete).
    * The normal return of this method indicates that the lock request (or lack thereof)
    * completed successfully, and that the current context now holds a lock of the requested
    * type.
    *
    * @param   lockType
    *          Type of lock to be obtained.  <code>LockType.NONE</code> is used to release an
    *          existing lock, and to continue with no lock.
    * @param   table
    *          Table name.
    * @param   update
    *          <code>true</code> if the lock state for the current table should be updated;
    *          else <code>false</code>.
    *
    * @throws  LockUnavailableException
    *          if a "no-wait" lock cannot be acquired immediately.
    *
    * @return  previous lock type.
    */
   public LockType lockTable(LockType lockType, String table, boolean update)
   throws LockUnavailableException
   {
      int type = LockType.toInt(lockType);
      return multiplexer.lockTable(mpID, type, table, update);
   }

   /**
    * Query the lock type currently held by the current context for the
    * specified database record.
    * <p>
    * This method should always return the <b>non-</b><code>NO_WAIT</code>
    * variants of a lock type.  That is, even if the actual lock type is
    * <code>LockType.EXCLUSIVE_NO_WAIT</code> or
    * <code>LockType.SHARE_NO_WAIT</code>, the simpler types of
    * <code>LockType.EXCLUSIVE</code> and <code>LockType.SHARE</code>,
    * respectively, should be returned.  The lack of a lock is required to
    * return <code>LockType.NONE</code> rather than <code>null</code>.
    *
    * @param   ident
    *          ID which uniquely identifies the record being queried.
    *
    * @return  The lock type currently held by the current context, never
    *          <code>null</code>.
    */
   public LockType queryLock(RecordIdentifier ident)
   {
      int type = multiplexer.queryLock(mpID, ident);
      
      return LockType.fromInt(type);
   }
}
