/*
** Module   : RemotePersistence.java
** Abstract : Persistence subclass which delegates some services to a remote P2J server
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20071018   @35556 Created initial version. Persistence subclass
**                           which delegates some services to a remote
**                           server.
** 002 ECF 20071115   @35912 Completed initial implementation. Obtain
**                           network proxies for LockManagerMultiplexer
**                           and DirtyReadMultiplexer to enable remote
**                           lock management and dirty read management.
** 003 SVL 20080111   @36771 Unnecessary parameter 'database' removed from
**                           createLockManager() and
**                           createDirtyReadManager() functions.
**                           createIdentityManager() function added.
** 004 ECF 20080611   @38707 Removed all references to DirtyReadManager
**                           related classes. This hierarchy has been
**                           obsoleted by the persist.dirty package.
** 005 ECF 20080806   @39308 Modified to conform to new IdentityManager API.
** 006 ECF 20131028          Import change.
*/

package com.goldencode.p2j.persist.remote;

import org.hibernate.dialect.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.id.*;
import com.goldencode.p2j.persist.lock.*;

/**
 * A subclass of {@link Persistence} which is intended to be used with a
 * database instance managed by a remote P2J server.  Most work is executed
 * directly against the database via Hibernate;  however, certain services
 * must be serialized through the remote P2J server instance.  These include:
 * <ul>
 *   <li><b>Record lock management</b>.  Record locking must be serialized
 *       through the P2J server instance which is authoritative for the target
 *       database, to ensure that the lock manager implementation in use for a
 *       given database can properly manage concurrent locking requests.  This
 *       is accomplished by instantiating a specialized {@link
 *       com.goldencode.p2j.persist.lock.LockManager LockManager} implementation
 *       ({@link RemoteLockManager}, which delegates requests for locking
 *       services to the remote server.
 *   <li><b>Primary key generation</b>.  Likewise, the management of new
 *       primary keys must be delegated to the remote server which is
 *       authoritative for the target database.  This is accomplished by
 *       instantiating a specialized {@link
 *       com.goldencode.p2j.persist.id.IdentityManager IdentityManager}
 *       implementation ({@link RemoteIdentityManager}).
 * </ul>
 * 
 * @author ECF
 */
public final class RemotePersistence
extends Persistence
{
   /** Network session with remote server */
   private final Session session;
   
   /**
    * Construct an instance for a particular, physical database, storing the
    * dialect class which Hibernate uses to create a portability layer.
    *
    * @param   database
    *          Database with which this instance is permanently associated.
    * @param   dialect
    *          Portability object used by Hibernate to issue commands to the
    *          database.
    */
   public RemotePersistence(Database database, Dialect dialect)
   {
      super(database, dialect);
      
      session = ConnectionManager.getSession(database);
   }
   
   /**
    * Create and initialize the lock manager to be used with this instance of
    * persistence services.
    *
    * @return  Lock manager instance.
    */
   protected LockManager createLockManager()
   {
      // Get lock manager multiplexer proxy.
      LockManagerMultiplexer multiplexer = (LockManagerMultiplexer)
         RemoteObject.obtainNetworkInstance(LockManagerMultiplexer.class,
                                            session);
      LockManager mgr = new RemoteLockManager(multiplexer);
      mgr.setDatabase(getDatabase());
      
      return mgr;
   }
   
   /**
    * Create and initialize the identity manager to be used with this instance
    * of persistence services.
    *
    * @return  Identity manager instance.
    */
   protected IdentityManager createIdentityManager()
   {
      // Get identity manager multiplexer proxy.
      IdentityManagerMultiplexer multiplexer = (IdentityManagerMultiplexer)
         RemoteObject.obtainNetworkInstance(IdentityManagerMultiplexer.class,
                                            session);

      IdentityManager mgr = new RemoteIdentityManager(multiplexer);
      mgr.setPersistence(this);

      return mgr;
   }
}
